"""Various type definitions used throughout stupidb."""

from numbers import Real
from typing import Callable, Hashable, Optional, Tuple, TypeVar

from .protocols import AdditiveWithInverse, Comparable
from .row import AbstractRow

Input = TypeVar("Input")
Input1 = TypeVar("Input1")
Input2 = TypeVar("Input2")
Input3 = TypeVar("Input3")
Output = TypeVar("Output")

T = TypeVar("T")
R = TypeVar("R", bound=Real)
R1 = TypeVar("R1", bound=Real)
R2 = TypeVar("R2", bound=Real)

OrderingKey = Tuple[Comparable[T], ...]
PartitionKey = Tuple[Tuple[str, Hashable], ...]

PartitionBy = Callable[[AbstractRow], Hashable]
OrderBy = Callable[[AbstractRow], Comparable[T]]
Preceding = Callable[[AbstractRow], AdditiveWithInverse[T]]
Following = Callable[[AbstractRow], AdditiveWithInverse[T]]

JoinPredicate = Callable[[AbstractRow, AbstractRow], Optional[bool]]
Predicate = Callable[[AbstractRow], Optional[bool]]

Projector = Callable[[AbstractRow], AbstractRow]
Result = TypeVar("Result")
Getter = Callable[[AbstractRow], T]



"""StupiDB user-facing API.

.. note::

   The join functions all take `right`, `predicate` and then `left` as
   arguments, **in that order**.

   This is intentional, and is the way the functions must be written to enable
   `currying <https://en.wikipedia.org/wiki/Currying>`_.  Currying is the
   technique that allows us to use the right shift operator (``>>``) to chain
   operations.

"""

from __future__ import annotations

import inspect
import operator
from typing import Any, Callable, Iterable, Mapping

import tabulate
import toolz
from public import private, public

from .aggregation import (
    AggregateSpecification,
    FrameClause,
    Nulls,
    WindowAggregateSpecification,
)
from .core import (
    Aggregation,
    CrossJoin,
    Difference,
    DifferenceAll,
    GroupBy,
    InnerJoin,
    Intersect,
    IntersectAll,
    Join,
    JoinPredicate,
    LeftJoin,
    Limit,
    Mutate,
    PartitionBy,
    Predicate,
    Projection,
    Projector,
    Relation,
    RightJoin,
    Selection,
    SortBy,
    Table,
    Union,
    UnionAll,
)
from .functions.associative import (
    Count,
    Max,
    Mean,
    Min,
    PopulationCovariance,
    PopulationStandardDeviation,
    PopulationVariance,
    SampleCovariance,
    SampleStandardDeviation,
    SampleVariance,
    Sum,
    Total,
)
from .functions.navigation import First, Lag, Last, Lead, Nth
from .functions.ranking import DenseRank, Rank, RowNumber
from .protocols import Comparable
from .row import AbstractRow
from .typehints import R1, R2, OrderBy, R, T


@public  # type: ignore[misc]
def const(x: T | None) -> Callable[[AbstractRow], T | None]:
    """Return a function that returns `x` regardless of input."""
    return lambda _: x


@public  # type: ignore[misc]
def get(name: str) -> Callable[[AbstractRow], T | None]:
    """Return a function that gets the `name` field from a row."""
    return operator.itemgetter(name)


@private  # type: ignore[misc]
class shiftable(toolz.curry):
    """Shiftable curry."""

    @property
    def __signature__(self) -> inspect.Signature:
        return inspect.signature(self.func)  # pragma: no cover

    def __rrshift__(self, other: Relation) -> shiftable:
        return self(other)


@public  # type: ignore[misc]
@shiftable
def table(rows: Iterable[Mapping[str, Any]]) -> Table:
    """Construct a relation from an iterable of mappings.

    Parameters
    ----------
    rows
        An iterable of mappings whose keys are :class:`str` instances.

    Examples
    --------
    >>> from stupidb import table
    >>> rows = [
    ...     dict(name="Bob", balance=-300),
    ...     dict(name="Bob", balance=-100),
    ...     dict(name="Alice", balance=400),
    ...     dict(name="Alice", balance=700),
    ... ]
    >>> t = table(rows)
    >>> t  # doctest: +ELLIPSIS
    name      balance
    ------  ---------
    Bob          -300
    Bob          -100
    Alice         400
    Alice         700

    """
    return Table.from_iterable(rows)


@public  # type: ignore[misc]
@shiftable
def cross_join(right: Relation, left: Relation) -> Join:
    """Return the Cartesian product of tuples from `left` and `right`.

    Parameters
    ----------
    right
        A relation
    left
        A relation

    Examples
    --------
    >>> from stupidb import cross_join, select, table
    >>> rows = [
    ...     dict(name="Bob", balance=-300),
    ...     dict(name="Bob", balance=-100),
    ...     dict(name="Alice", balance=400),
    ...     dict(name="Alice", balance=700),
    ... ]
    >>> t = table(rows)
    >>> s = table(rows)
    >>> crossed = cross_join(t, s)
    >>> crossed >> select(
    ...     left_name=lambda row: row.left["name"],
    ...     left_balance=lambda row: row.left["balance"],
    ...     right_name=lambda row: row.right["name"],
    ...     right_balance=lambda row: row.right["balance"]
    ... )
    left_name      left_balance  right_name      right_balance
    -----------  --------------  ------------  ---------------
    Bob                    -300  Bob                      -300
    Bob                    -300  Bob                      -100
    Bob                    -300  Alice                     400
    Bob                    -300  Alice                     700
    Bob                    -100  Bob                      -300
    Bob                    -100  Bob                      -100
    Bob                    -100  Alice                     400
    Bob                    -100  Alice                     700
    Alice                   400  Bob                      -300
    Alice                   400  Bob                      -100

    """
    return CrossJoin(left, right)


@public  # type: ignore[misc]
@shiftable
def inner_join(right: Relation, predicate: JoinPredicate, left: Relation) -> Join:
    """Join `left` and `right` relations using `predicate`.

    Drop rows if `predicate` returns ``False``.

    Parameters
    ----------
    right
        A relation
    predicate
        A callable taking two arguments and returning a :class:`bool`.

    Examples
    --------
    >>> from stupidb import inner_join, table
    >>> rows = [
    ...     dict(name="Bob", balance=-300),
    ...     dict(name="Bob", balance=-100),
    ...     dict(name="Alice", balance=400),
    ...     dict(name="Alice", balance=700),
    ... ]
    >>> t = table(rows)
    >>> s = table(rows)
    >>> t >> inner_join(
    ...     s,
    ...     lambda left, right: left["balance"] < right["balance"]
    ... ) >> select(name=lambda r: r.left["name"], bal=lambda r: r.left["balance"])
    name      bal
    ------  -----
    Bob      -300
    Bob      -300
    Bob      -300
    Bob      -100
    Bob      -100
    Alice     400

    """
    return InnerJoin(left, right, predicate)


@public  # type: ignore[misc]
@shiftable
def left_join(right: Relation, predicate: JoinPredicate, left: Relation) -> LeftJoin:
    """Join `left` and `right` relations using `predicate`.

    Drop rows if `predicate` returns ``False``.  Returns at least one of every
    row from `left`.

    Parameters
    ----------
    right
        A relation
    predicate
        A callable taking two arguments and returning a :class:`bool`.

    """
    return LeftJoin(left, right, predicate)


@public  # type: ignore[misc]
@shiftable
def right_join(right: Relation, predicate: JoinPredicate, left: Relation) -> RightJoin:
    """Join `left` and `right` relations using `predicate`.

    Drop rows if `predicate` returns ``False``.  Returns at least one of every
    row from `right`.

    Parameters
    ----------
    right
        A relation
    predicate
        A callable taking two arguments and returning a :class:`bool`.

    """
    return RightJoin(left, right, predicate)


@public  # type: ignore[misc]
@shiftable
def full_join(right: Relation, predicate: JoinPredicate, left: Relation) -> Relation:
    """Full outer join."""
    raise NotImplementedError("full outer joins are not yet supported")


@private  # type: ignore[misc]
@shiftable
def _order_by(order_by: tuple[OrderBy, ...], nulls: Nulls, child: Relation) -> SortBy:
    return SortBy(child, order_by, nulls)


@public  # type: ignore[misc]
def order_by(*order_by: OrderBy, nulls: Nulls = Nulls.FIRST) -> SortBy:
    """Order the rows of the child operator according to `order_by`.

    Parameters
    ----------
    order_by
        A sequence of ``OrderBy`` instances
    nulls
        One of :class:`~stupidb.aggregation.Nulls` indicating how to treat
        nulls when sorting. :attr:`~stupidb.aggregation.Nulls.FIRST` treats
        nulls as less than every other value, and
        :attr:`~stupidb.aggregation.Nulls.LAST` treats them as greater than
        every other value.

    Examples
    --------
    >>> from stupidb import order_by, table
    >>> rows = [
    ...     dict(name="Bob", balance=-300),
    ...     dict(name="Alice", balance=400),
    ...     dict(name="Bob", balance=-100),
    ...     dict(name="Alice", balance=700),
    ... ]
    >>> ordered = table(rows) >> order_by(lambda r: r.balance)
    >>> balances = [row.balance for row in ordered]
    >>> balances
    [-300, -100, 400, 700]

    """
    return _order_by(order_by, nulls)


@private  # type: ignore[misc]
@shiftable
def _select(
    projectors: Mapping[str, Projector | WindowAggregateSpecification], child: Relation
) -> Projection:
    return Projection(child, projectors)


@public  # type: ignore[misc]
def select(**projectors: Projector | WindowAggregateSpecification) -> Projection:
    """Subset or compute new columns from `projectors`.

    Parameters
    ----------
    projectors
        A mapping from :class:`str` to :data:`FullProjector` instances.

    Examples
    --------
    >>> from stupidb import select, table
    >>> rows = [
    ...     dict(name="Bob", balance=-300),
    ...     dict(name="Alice", balance=400),
    ...     dict(name="Bob", balance=-100),
    ...     dict(name="Alice", balance=700),
    ... ]
    >>> names = table(rows) >> select(lower_name=lambda r: r.name.lower())
    >>> names = [row.lower_name for row in names]
    >>> names
    ['bob', 'alice', 'bob', 'alice']

    See Also
    --------
    mutate

    """
    valid_projectors = {
        name: projector
        for name, projector in projectors.items()
        if callable(projector) or isinstance(projector, WindowAggregateSpecification)
    }
    if len(valid_projectors) != len(projectors):
        raise TypeError("Invalid projection")
    return _select(projectors)


@private  # type: ignore[misc]
@shiftable
def _mutate(
    mutators: Mapping[str, Projector | WindowAggregateSpecification], child: Relation
) -> Mutate:
    return Mutate(child, mutators)


@public  # type: ignore[misc]
def mutate(**mutators: Projector | WindowAggregateSpecification) -> Mutate:
    """Add new columns specified by `mutators`.

    Parameters
    ----------
    projectors
        A mapping from :class:`str` to :data:`FullProjector` instances.

    Notes
    -----
    Columns are appended, unlike :func:`~stupidb.api.select`.

    Examples
    --------
    >>> from stupidb import mutate, table
    >>> rows = [
    ...     dict(name="Bob", balance=-300),
    ...     dict(name="Alice", balance=400),
    ...     dict(name="Bob", balance=-100),
    ...     dict(name="Alice", balance=700),
    ... ]
    >>> rows = table(rows) >> mutate(lower_name=lambda r: r.name.lower())
    >>> rows
    name      balance  lower_name
    ------  ---------  ------------
    Bob          -300  bob
    Alice         400  alice
    Bob          -100  bob
    Alice         700  alice

    See Also
    --------
    select

    """
    return _mutate(mutators)


@public  # type: ignore[misc]
@shiftable
def sift(predicate: Predicate, child: Relation) -> Selection:
    """Filter rows in `child` according to `predicate`.

    Parameters
    ----------
    predicate
        A callable of one argument taking an :class:`~stupidb.row.AbstractRow`
        and returning a ``bool``.

    Examples
    --------
    >>> from stupidb import sift, table
    >>> rows = [
    ...     dict(name="Bob", balance=-300),
    ...     dict(name="Alice", balance=400),
    ...     dict(name="Bob", balance=-100),
    ...     dict(name="Alice", balance=700),
    ... ]
    >>> rows = table(rows) >> sift(lambda r: r.name.lower().startswith("a"))
    >>> rows
    name      balance
    ------  ---------
    Alice         400
    Alice         700

    """
    return Selection(child, predicate)


@public  # type: ignore[misc]
def exists(relation: Relation) -> bool:
    """Compute whether any of the rows in `relation` are truthy.

    This is useful for computing semi-joins.

    """
    return any(relation)


@private  # type: ignore[misc]
@shiftable
def _aggregate(
    aggregations: Mapping[str, AggregateSpecification], child: Relation
) -> Aggregation:
    return Aggregation(child, aggregations)


@public  # type: ignore[misc]
def aggregate(**aggregations: AggregateSpecification) -> Aggregation:
    """Aggregate values from the child operator using `aggregations`.

    Parameters
    ----------
    aggregations
        A mapping from :class:`str` column names to
        :class:`~stupidb.aggregation.AggregateSpecification` instances.

    Examples
    --------
    Compute the average of a column:

    >>> from stupidb import aggregate, group_by, mean, table
    >>> rows = [
    ...     dict(name="Bob", age=30, timezone="America/New_York"),
    ...     dict(name="Susan", age=20, timezone="America/New_York"),
    ...     dict(name="Joe", age=41, timezone="America/Los_Angeles"),
    ...     dict(name="Alice", age=39, timezone="America/Los_Angeles"),
    ... ]
    >>> average_age = table(rows) >> aggregate(avg_age=mean(lambda r: r.age))
    >>> average_age
      avg_age
    ---------
         32.5

    Compute the average a column, grouped by another column:

    >>> average_age_by_timezone = (
    ...     table(rows) >> group_by(tz=lambda r: r.timezone)
    ...                 >> aggregate(avg_age=mean(lambda r: r.age))
    ... )
    >>> average_age_by_timezone
    tz                     avg_age
    -------------------  ---------
    America/New_York            25
    America/Los_Angeles         40

    See Also
    --------
    group_by

    """
    return _aggregate(aggregations)


@public  # type: ignore[misc]
@shiftable
def over(
    window: FrameClause, child: AggregateSpecification
) -> WindowAggregateSpecification:
    """Construct a window aggregate.

    Parameters
    ----------
    window
        A :class:`~stupidb.aggregation.FrameClause` instance constructed from
        :class:`~stupidb.aggregation.Window.rows` or
        :class:`~stupidb.aggregation.Window.range`.
    child
        The aggregation to compute over `window`

    Notes
    -----
    This is one of the few user-facing functions that does **not** return a
    :class:`~stupidb.core.Relation`. The behavior of materializing the rows
    of the result of calling this function is undefined.

    Examples
    --------
    >>> from stupidb import Window, over, mean, select, table
    >>> from datetime import date, timedelta
    >>> today = date(2019, 2, 9)
    >>> days = timedelta(days=1)
    >>> rows = [
    ...     {"name": "Alice", "balance": 400, "date": today},
    ...     {"name": "Alice", "balance": 300, "date": today + 1 * days},
    ...     {"name": "Alice", "balance": 100, "date": today + 2 * days},
    ...     {"name": "Bob", "balance": -150, "date": today - 4 * days},
    ...     {"name": "Bob", "balance": 200, "date": today - 3 * days},
    ... ]
    >>> t = table(rows)
    >>> window = Window.range(
    ...     partition_by=[lambda r: r.name],
    ...     order_by=[lambda r: r.date],
    ...     preceding=lambda r: 2 * days  # two days behind + the current row
    ... )
    >>> avg_balance_per_person = table(rows) >> select(
    ...     name=lambda r: r.name,
    ...     avg_balance=mean(lambda r: r.balance) >> over(window),
    ...     balance=lambda r: r.balance,
    ...     date=lambda r: r.date,
    ... ) >> order_by(lambda r: r.name, lambda r: r.date)
    >>> avg_balance_per_person
    name      balance  date          avg_balance
    ------  ---------  ----------  -------------
    Alice         400  2019-02-09        400
    Alice         300  2019-02-10        350
    Alice         100  2019-02-11        266.667
    Bob          -150  2019-02-05       -150
    Bob           200  2019-02-06         25

    """
    return WindowAggregateSpecification(child.aggregate_type, child.getters, window)


@private  # type: ignore[misc]
@shiftable
def _group_by(group_by: Mapping[str, PartitionBy], child: Relation) -> GroupBy:
    return GroupBy(child, group_by)


@public  # type: ignore[misc]
def group_by(**group_by: PartitionBy) -> GroupBy:
    """Group the rows of the child operator according to `group_by`.

    Parameters
    ----------
    group_by
        A mapping of :class:`str` column names to functions that compute
        grouping keys.

    Notes
    -----
    Iterating over the rows of the result of this function is not very useful,
    since its :meth:`~stupidb.core.GroupBy.__iter__` method just yields
    the rows of its child. A call to this function is best followed by a call
    to :func:`~stupidb.api.aggregate`.

    Examples
    --------
    >>> from stupidb import aggregate, group_by, mean, table
    >>> rows = [
    ...     dict(name="Bob", age=30, timezone="America/New_York"),
    ...     dict(name="Susan", age=20, timezone="America/New_York"),
    ...     dict(name="Joe", age=41, timezone="America/Los_Angeles"),
    ...     dict(name="Alice", age=39, timezone="America/Los_Angeles"),
    ... ]
    >>> average_age_by_timezone = (
    ...     table(rows) >> group_by(tz=lambda r: r.timezone)
    ...                 >> aggregate(avg_age=mean(lambda r: r.age))
    ... )
    >>> average_age_by_timezone
    tz                     avg_age
    -------------------  ---------
    America/New_York            25
    America/Los_Angeles         40

    See Also
    --------
    aggregate

    """
    return _group_by(group_by)


@public  # type: ignore[misc]
@shiftable
def union(right: Relation, left: Relation) -> Union:
    """Compute the union of `left` and `right`, ignoring duplicate rows.

    Parameters
    ----------
    right
         A relation
    left
         A relation

    See Also
    --------
    union_all

    """
    return Union(left, right)


@public  # type: ignore[misc]
@shiftable
def union_all(right: Relation, left: Relation) -> UnionAll:
    """Compute the union of `left` and `right`, preserving duplicate rows.

    Parameters
    ----------
    right
         A relation
    left
         A relation

    See Also
    --------
    union

    """
    return UnionAll(left, right)


@public  # type: ignore[misc]
@shiftable
def intersect(right: Relation, left: Relation) -> Intersect:
    """Compute the intersection of `left` and `right`, ignoring duplicate rows.

    Parameters
    ----------
    right
         A relation
    left
         A relation

    See Also
    --------
    intersect_all

    """
    return Intersect(left, right)


@public  # type: ignore[misc]
@shiftable
def intersect_all(right: Relation, left: Relation) -> IntersectAll:
    """Compute the intersection of `left` and `right`, preserving duplicates.

    Parameters
    ----------
    right
         A relation
    left
         A relation

    See Also
    --------
    intersect

    """
    return IntersectAll(left, right)


@public  # type: ignore[misc]
@shiftable
def difference(right: Relation, left: Relation) -> Difference:
    """Compute the set difference of `left` and `right`.

    Parameters
    ----------
    right
         A relation
    left
         A relation

    """
    return Difference(left, right)


@public  # type: ignore[misc]
@shiftable
def difference_all(right: Relation, left: Relation) -> DifferenceAll:
    """Compute the set difference of `left` and `right`, preserving duplicates.

    Parameters
    ----------
    right
         A relation
    left
         A relation

    """
    return DifferenceAll(left, right)


@public  # type: ignore[misc]
@shiftable
def limit(limit: int | None, relation: Relation, *, offset: int = 0) -> Limit:
    """Return the rows in `relation` starting from `offset` up to `limit`.

    Parameters
    ----------
    limit
        The number of rows starting from offset to produce
    relation
        Relation whose rows to limit
    offset
        The number of rows to skip before yielding

    """
    if offset < 0:
        raise ValueError(f"invalid offset, must be non-negative: {offset}")
    if limit is not None and limit < 0:
        raise ValueError(f"invalid limit, must be non-negative or None: {limit}")
    return Limit(relation, offset=offset, limit=limit)


@public  # type: ignore[misc]
def count(x: Callable[[AbstractRow], T | None]) -> AggregateSpecification:
    """Count the number of non-NULL values of `x`.

    Parameters
    ----------
    x
        A column getter.

    """
    return AggregateSpecification(Count, x)


@public  # type: ignore[misc]
def sum(x: Callable[[AbstractRow], R | None]) -> AggregateSpecification:
    """Compute the sum of `x`, with an empty column summing to NULL.

    Parameters
    ----------
    x
        A column getter.

    """
    return AggregateSpecification(Sum, x)


@public  # type: ignore[misc]
def total(x: Callable[[AbstractRow], R | None]) -> AggregateSpecification:
    """Compute the sum of `x`, with an empty column summing to zero.

    Parameters
    ----------
    x
        A column getter.

    """
    return AggregateSpecification(Total, x)


@public  # type: ignore[misc]
def first(x: Callable[[AbstractRow], T | None]) -> AggregateSpecification:
    """Compute the first row of `x` over a window.

    Parameters
    ----------
    x
        A column getter.

    """
    return AggregateSpecification(First, x)


@public  # type: ignore[misc]
def last(x: Callable[[AbstractRow], T | None]) -> AggregateSpecification:
    """Compute the last row of `x` over a window.

    Parameters
    ----------
    x
        A column getter.

    """
    return AggregateSpecification(Last, x)


@public  # type: ignore[misc]
def nth(
    x: Callable[[AbstractRow], T | None],
    i: Callable[[AbstractRow], int | None],
) -> AggregateSpecification:
    """Compute the `i`-th row of `x` over a window.

    Parameters
    ----------
    x
        Column selector.
    i
        Callable to compute the row offset of the window to return.

    """
    return AggregateSpecification(Nth, x, i)


@public  # type: ignore[misc]
def row_number() -> AggregateSpecification:
    """Compute the row number over a window."""
    return AggregateSpecification(RowNumber)


@public  # type: ignore[misc]
def rank() -> AggregateSpecification:
    """Rank the rows of a relation based on the ordering key given in `over`."""
    return AggregateSpecification(Rank)


@public  # type: ignore[misc]
def dense_rank() -> AggregateSpecification:
    """Rank the rows of a relation based on the ordering key given in `over`."""
    return AggregateSpecification(DenseRank)


@public  # type: ignore[misc]
def lead(
    x: Callable[[AbstractRow], T | None],
    n: Callable[[AbstractRow], int | None] = const(1),
    default: Callable[[AbstractRow], T | None] = const(None),
) -> AggregateSpecification:
    """Lead a column `x` by `n` rows, using `default` for NULL values.

    Parameters
    ----------
    x
        A column selector.
    n
        A callable computing the number of rows to lead. Defaults to a lead of
        **1** row. The callable takes the current row as input and thus the
        lead can be computed relative to the current row.
    default
        A callable computing the default value for the lead if the row would
        produce a NULL value when led. The callable takes the current row as
        input and thus the default can be computed relative to the current row.

    """
    return AggregateSpecification(Lead, x, n, default)


@public  # type: ignore[misc]
def lag(
    x: Callable[[AbstractRow], T | None],
    n: Callable[[AbstractRow], int | None] = const(1),
    default: Callable[[AbstractRow], T | None] = const(None),
) -> AggregateSpecification:
    """Lag a column `x` by `n` rows, using `default` for NULL values.

    Parameters
    ----------
    x
        A column selector.
    n
        A callable computing the number of rows to lag. Defaults to a lag of
        **1** row. The callable takes the current row as input and thus the lag
        can be computed relative to the current row.
    default
        A callable computing the default value for the lag if the row would
        produce a NULL value when lagged. The callable takes the current row as
        input and thus the default can be computed relative to the current row.

    """
    return AggregateSpecification(Lag, x, n, default)


@public  # type: ignore[misc]
def mean(x: Callable[[AbstractRow], R | None]) -> AggregateSpecification:
    """Compute the average of a column.

    Parameters
    ----------
    x
        A column selector.

    """
    return AggregateSpecification(Mean, x)


@public  # type: ignore[misc]
def min(x: Callable[[AbstractRow], Comparable | None]) -> AggregateSpecification:
    """Compute the minimum of a column.

    Parameters
    ----------
    x
        A column selector.

    """
    return AggregateSpecification(Min, x)


@public  # type: ignore[misc]
def max(x: Callable[[AbstractRow], Comparable | None]) -> AggregateSpecification:
    """Compute the maximum of a column.

    Parameters
    ----------
    x
        A column selector.

    """
    return AggregateSpecification(Max, x)


@public  # type: ignore[misc]
def cov_samp(
    x: Callable[[AbstractRow], R1 | None], y: Callable[[AbstractRow], R2 | None]
) -> AggregateSpecification:
    """Compute the sample covariance of two columns.

    Parameters
    ----------
    x
        A column selector.
    y
        A column selector.

    """
    return AggregateSpecification(SampleCovariance, x, y)


@public  # type: ignore[misc]
def var_samp(x: Callable[[AbstractRow], R | None]) -> AggregateSpecification:
    """Compute the sample variance of a column.

    Parameters
    ----------
    x
        A column selector.

    """
    return AggregateSpecification(SampleVariance, x)


@public  # type: ignore[misc]
def stdev_samp(x: Callable[[AbstractRow], R | None]) -> AggregateSpecification:
    """Compute the sample standard deviation of a column.

    Parameters
    ----------
    x
        A column selector.

    """
    return AggregateSpecification(SampleStandardDeviation, x)


@public  # type: ignore[misc]
def cov_pop(
    x: Callable[[AbstractRow], R1 | None], y: Callable[[AbstractRow], R2 | None]
) -> AggregateSpecification:
    """Compute the population covariance of two columns.

    Parameters
    ----------
    x
        A column selector.
    y
        A column selector.

    """
    return AggregateSpecification(PopulationCovariance, x, y)


@public  # type: ignore[misc]
def var_pop(x: Callable[[AbstractRow], R]) -> AggregateSpecification:
    """Compute the population variance of a column.

    Parameters
    ----------
    x
        A column selector.

    """
    return AggregateSpecification(PopulationVariance, x)


@public  # type: ignore[misc]
def stdev_pop(x: Callable[[AbstractRow], R | None]) -> AggregateSpecification:
    """Compute the population standard deviation of a column.

    Parameters
    ----------
    x
        A column selector.

    """
    return AggregateSpecification(PopulationStandardDeviation, x)


@public  # type: ignore[misc]
@shiftable
def pretty(
    rows: Relation,
    *,
    n: int | None = 10,
    tablefmt: str = "simple",
    headers: str = "keys",
    **kwargs: Any,
) -> str:
    """Pretty-format a relation.

    Parameters
    ----------
    rows
        The relation to print
    n
        The number of rows to format; A value of `None` will pretty format all
        rows
    tablefmt
        The kind of table to use for formatting
    headers
        A string indicating how to compute column names
    kwargs
        Additional keyword arguments passed to the `tabulate.tabulate`
        function

    Returns
    -------
    str
        Pretty-formatted relation

    See Also
    --------
    stupidb.api.show

    """
    return tabulate.tabulate(
        limit(n, rows), tablefmt=tablefmt, headers=headers, **kwargs
    )


@public  # type: ignore[misc]
@shiftable
def show(rows: Relation, **kwargs: Any) -> None:
    """Pretty-print a relation.

    Parameters
    ----------
    rows
        The relation to print
    kwargs
        Additional keyword arguments passed to the `stupidb.api.pretty`
        function

    See Also
    --------
    stupidb.api.pretty

    """
    print(pretty(rows, **kwargs))



from __future__ import annotations

import sqlite3
from datetime import date
from typing import Mapping, Sequence, Union

import pytest

from stupidb.api import table
from stupidb.core import Table

Element = Union[date, float, int, str, None]


@pytest.fixture(scope="session")  # type: ignore[misc]
def rows() -> list[dict[str, Element]]:
    return [
        dict(z="a", a=1, b=2, e=1),
        dict(z="b", a=2, b=-1, e=2),
        dict(z="a", a=3, b=4, e=3),
        dict(z="a", a=4, b=-3, e=4),
        dict(z="a", a=1, b=-3, e=5),
        dict(z="b", a=2, b=-3, e=6),
        dict(z="b", a=3, b=-3, e=7),
    ]


@pytest.fixture(scope="session")  # type: ignore[misc]
def left(rows: list[dict[str, Element]]) -> list[dict[str, Element]]:
    return rows


@pytest.fixture(scope="session")  # type: ignore[misc]
def right(rows: list[dict[str, Element]]) -> list[dict[str, Element]]:
    return [
        dict(z="a", a=1, b=2, e=1),
        dict(z="c", a=2, b=-1, e=2),
        dict(z="a", a=3, b=4, e=3),
        dict(z="c", a=4, b=-3, e=4),
        dict(z="a", a=1, b=-3, e=5),
        dict(z="c", a=2, b=-3, e=6),
        dict(z="c", a=3, b=-3, e=7),
    ]


def tupleize(row: Mapping[str, Element]) -> frozenset[tuple[str, Element]]:
    return frozenset(row.items())


def assert_rowset_equal(
    left: Sequence[Mapping[str, Element]],
    right: Sequence[Mapping[str, Element]],
) -> None:
    assert set(map(tupleize, left)) == set(map(tupleize, right))


@pytest.fixture(scope="session")  # type: ignore[misc]
def t_rows() -> list[dict[str, Element]]:
    return [
        dict(name="alice", date=date(2018, 1, 1), balance=2),
        dict(name="alice", date=date(2018, 1, 4), balance=4),
        dict(name="alice", date=date(2018, 1, 6), balance=-3),
        dict(name="alice", date=date(2018, 1, 7), balance=-3),
        dict(name="bob", date=date(2018, 1, 2), balance=-1),
        dict(name="bob", date=date(2018, 1, 3), balance=-3),
        dict(name="bob", date=date(2018, 1, 4), balance=-3),
    ]


@pytest.fixture(scope="session")  # type: ignore[misc]
def t_table(t_rows: list[dict[str, Element]]) -> Table:
    return table(t_rows)


@pytest.fixture(scope="session")  # type: ignore[misc]
def employee() -> list[dict[str, Element]]:
    return [
        dict(last_name="Rafferty", department_id=1),
        dict(last_name="Jones", department_id=2),
        dict(last_name="Heisenberg", department_id=2),
        dict(last_name="Robinson", department_id=3),
        dict(last_name="Smith", department_id=3),
        dict(last_name="Williams", department_id=None),
    ]


@pytest.fixture(scope="session")  # type: ignore[misc]
def department() -> list[dict[str, Element]]:
    return [
        dict(department_id=1, department_name="Sales"),
        dict(department_id=2, department_name="Engineering"),
        dict(department_id=3, department_name="Clerical"),
        dict(department_id=4, department_name="Marketing"),
    ]


@pytest.fixture(scope="session")  # type: ignore[misc]
def con(
    rows: list[dict[str, Element]],
    left: list[dict[str, Element]],
    right: list[dict[str, Element]],
    employee: list[dict[str, Element]],
    department: list[dict[str, Element]],
) -> sqlite3.Connection:
    connection = sqlite3.connect(":memory:")

    connection.execute("CREATE TABLE rows (z text, a integer, b integer, e integer)")
    connection.executemany(
        "INSERT INTO rows VALUES (?, ?, ?, ?)", (tuple(row.values()) for row in rows)
    )

    connection.execute("CREATE TABLE left (z text, a integer, b integer, e integer)")
    connection.executemany(
        "INSERT INTO left VALUES (?, ?, ?, ?)", (tuple(row.values()) for row in left)
    )

    connection.execute("CREATE TABLE right (z text, a integer, b integer, e integer)")
    connection.executemany(
        "INSERT INTO right VALUES (?, ?, ?, ?)", (tuple(row.values()) for row in right)
    )

    connection.execute(
        """
        CREATE TABLE department (
            department_id INTEGER PRIMARY KEY,
            department_name TEXT NOT NULL
        )
        """
    )
    connection.executemany(
        "INSERT INTO department VALUES (?, ?)",
        (tuple(row.values()) for row in department),
    )

    connection.execute(
        """
        CREATE TABLE employee (
            employee_id INTEGER PRIMARY KEY AUTOINCREMENT,
            last_name TEXT NOT NULL,
            department_id INTEGER,
            FOREIGN KEY (department_id) REFERENCES department (department_id)
        )
        """
    )
    connection.executemany(
        "INSERT INTO employee (last_name, department_id) VALUES (?, ?)",
        (tuple(row.values()) for row in employee),
    )

    return connection



"""Animate construction of segment trees."""

import argparse
import collections
import os
import subprocess
import sys
import tempfile
from typing import BinaryIO, Iterator, MutableMapping, MutableSequence, Sequence

import pydot

from ..functions.associative import Sum
from . import indextree
from .bitgraph import BitGraph
from .bitset import BitSet
from .segmenttree import SegmentTree


class SegmentTreeAnimator:
    """Animate construction of a segment tree."""

    __slots__ = ("segment_tree",)

    def __init__(self, segment_tree: SegmentTree) -> None:
        """Construct a instance of a segment tree animation class."""
        self.segment_tree = segment_tree

    def make_graph(self, *, font: str) -> pydot.Dot:
        """Return a graph created from ``self.nodes``."""
        segment_tree = self.segment_tree
        tree = indextree.IndexTree(
            height=segment_tree.height, fanout=segment_tree.fanout
        )
        leaves = tree.leaves
        seen = BitSet()
        queue = collections.deque(leaves)

        graph = pydot.Dot(graph_type="digraph")
        for node in tree.nodes:
            graph.add_node(pydot.Node(node))

        while queue:
            node = queue.popleft()
            if node not in seen:
                seen.add(node)
                parent = tree.parent(node)
                if parent != node:
                    graph.add_edge(pydot.Edge(parent, node, dir="back"))
                queue.append(parent)
                if node in leaves:
                    result = segment_tree.nodes[node].finalize()
                    label = "" if result is None else str(result)
                else:
                    label = ""
                (pydot_node,) = graph.get_node(str(node))
                pydot_node.set_label(label)
                pydot_node.set_fontcolor("black")
                pydot_node.set_fontname(f"{font} bold")
                pydot_node.set_fillcolor("white")
                pydot_node.set_style("filled")

        return graph

    def iterframes(self, *, font: str) -> Iterator[bytes]:
        """Produce the frames of an animated construction of the tree."""
        graph = self.make_graph(font=font)
        segment_tree = self.segment_tree

        bit_graph = BitGraph.from_vertices_and_edges(
            vertices=(int(node.get_name()) for node in graph.get_nodes()),
            edges=(
                (edge.get_source(), edge.get_destination())
                for edge in graph.get_edges()
            ),
        )

        queue = collections.deque(bit_graph.in_edges)
        parent_count: MutableMapping[int, int] = collections.Counter()
        nodes = [segment_tree.aggregate_type() for _ in range(len(segment_tree.nodes))]
        for leaf in queue:
            nodes[leaf] = segment_tree.nodes[leaf]

        seen = BitSet()
        while queue:
            node = queue.popleft()
            if node not in seen:
                seen.add(node)
                parent = bit_graph.predecessors[node]
                node_agg = nodes[node]
                parent_agg = nodes[parent]
                parent_agg.combine(node_agg)
                parent_count[parent] += 1

                (pydot_node,) = graph.get_node(str(node))
                pydot_node.set_fillcolor("blue")
                pydot_node.set_fontcolor("white")
                pydot_node.set_fontname(f"{font} bold")

                result = parent_agg.finalize()

                (pydot_parent_node,) = graph.get_node(str(parent))
                pydot_parent_node.set_label("" if result is None else str(result))

                yield graph.create_gif()

                # don't traverse the root, since it will already contain its
                # full aggregate value due to the way we traverse
                if parent:
                    queue.append(parent)

                if parent_count[parent] == segment_tree.fanout - 1:
                    pydot_parent_node.set_fillcolor("red")
                    pydot_parent_node.set_fontcolor("white")
                    pydot_parent_node.set_fontname(f"{font} bold")
                    result = parent_agg.finalize()
                    pydot_parent_node.set_label("" if result is None else str(result))

    def animate(
        self, output: BinaryIO, font: str = "Helvetica", fps: float = 1.5
    ) -> None:
        """Convert this segment tree's construction into an animated gif.

        Parameters
        ----------
        output
            A writable binary output stream.
        font
            The font to use for text in the resulting animation.
        fps
            Frames per second.

        Examples
        --------
        >>> import os
        >>> from stupidb.associative.segmenttree import SegmentTree
        >>> from stupidb.functions.associative import Sum
        >>> from stupidb.associative.animate import SegmentTreeAnimator
        >>> segment_tree = SegmentTree([(1,), (2,), (3,), (4,)], Sum, fanout=2)
        >>> animator = SegmentTreeAnimator(segment_tree)
        >>> with open(os.devnull, "wb") as devnull:
        ...     animator.animate(devnull)

        """
        # NB: imagemagick units are 1/100th of a second
        centis_per_frame = 100.0 / fps
        assert centis_per_frame > 0, "frame duration is <= 0"

        with tempfile.TemporaryDirectory() as d:
            paths: MutableSequence[str] = []

            for i, frame in enumerate(self.iterframes(font=font)):
                path = os.path.join(d, str(i))
                with open(path, mode="wb") as f:
                    f.write(frame)
                paths.append(path)

            subprocess.run(
                [
                    "convert",
                    "-delay",
                    str(centis_per_frame),
                    "-loop",
                    "0",
                    *paths,
                    "gif:-",
                ],
                stdout=output,
            )


def parse_args() -> argparse.Namespace:
    p = argparse.ArgumentParser(
        description=(
            "Animate the construction of a segment tree build "
            "with the Sum aggregation."
        ),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    p.add_argument(
        "-o",
        "--output-file",
        type=argparse.FileType("wb"),
        default=sys.stdout.buffer,
        help="Where to write the animated GIF. Defaults to stdout.",
    )
    p.add_argument(
        "-f",
        "--fanout",
        type=int,
        default=2,
        help="Segment tree fanout.",
    )
    p.add_argument(
        "-l",
        "--leaf",
        type=int,
        action="append",
        default=[],
        help="The leaves of the segment tree.",
    )
    p.add_argument(
        "-r",
        "--frame-rate",
        type=float,
        default=1.5,
        help="Frames per second of the resulting animation.",
    )
    p.add_argument(
        "-F",
        "--font",
        type=str,
        default="Helvetica",
        help="Node font.",
    )
    return p.parse_args()


def main(
    *,
    output_file: BinaryIO,
    fanout: int,
    leaf: Sequence[int],
    frame_rate: float,
    font: str,
) -> None:
    """Animate the construction of a SegmentTree."""
    if not leaf:
        leaves = [(lf,) for lf in range(1, 9)]
    else:
        leaves = [(lf,) for lf in leaf]
    segment_tree: SegmentTree = SegmentTree(leaves, Sum, fanout=fanout)
    animator = SegmentTreeAnimator(segment_tree)
    animator.animate(output_file, font=font, fps=frame_rate)


if __name__ == "__main__":  # pragma: no cover
    args = parse_args()
    main(
        output_file=args.output_file,
        fanout=args.fanout,
        leaf=args.leaf,
        frame_rate=args.frame_rate,
        font=args.font,
    )



"""Various stupidb related protocol classes."""

import abc
from typing import Any, TypeVar

from typing_extensions import Protocol

B = TypeVar("B", contravariant=True)


class Comparable(Protocol[B]):
    """A protocol for comparable objects."""

    @abc.abstractmethod
    def __eq__(self, other: Any) -> bool:
        """Return whether `self` equals `other`."""

    def __ne__(self, other: Any) -> bool:
        """Return whether `self` does not equal `other`."""
        return not (self == other)

    @abc.abstractmethod
    def __lt__(self, other: B) -> bool:
        """Return whether `self` is less than `other`."""

    def __gt__(self, other: B) -> bool:
        """Return whether `self` is greater than `other`."""
        return (not self < other) and self != other

    def __le__(self, other: B) -> bool:
        """Return whether `self` is less than or equal to `other`."""
        return self < other or self == other

    def __ge__(self, other: B) -> bool:
        """Return whether `self` is greater than or equal to `other`."""
        return not self < other


A = TypeVar("A")


class AdditiveWithInverse(Protocol[A]):
    """A protocol for objects that are additive with an inverse."""

    @abc.abstractmethod
    def __add__(self, other: A) -> A:
        """Add `other` to `self`."""

    @abc.abstractmethod
    def __radd__(self, other: A) -> A:
        """Add `self` to `other`."""

    @abc.abstractmethod
    def __sub__(self, other: A) -> A:
        """Subtract `other` from `self`."""

    @abc.abstractmethod
    def __rsub__(self, other: A) -> A:
        """Subtract `self` from `other`."""

    @abc.abstractmethod
    def __neg__(self) -> A:
        """Negate `self`."""

    @abc.abstractmethod
    def __pos__(self) -> A:
        """Return positive `self`."""



"""An efficiently stored set of unsigned integers."""

from typing import Any, Iterable, Iterator, MutableSet


class BitSet(MutableSet[int]):
    """A efficiently stored set of unsigned integers."""

    __slots__ = "bits", "len"

    def __init__(self, bits: Iterable[int] = ()) -> None:
        """Construct a bitset."""
        self.bits = 0
        self.len = 0

        for bit in bits:
            self.add(bit)

    def __contains__(self, bit: Any) -> bool:
        """Check whether `bit` is in the set."""
        return (self.bits & (1 << bit)) != 0

    def __iter__(self) -> Iterator[int]:
        """Iterate over the elements of the set."""
        return filter(self.__contains__, range(self.bits.bit_length()))

    def __len__(self) -> int:
        """Return the number of set bits."""
        return self.len

    def __repr__(self) -> str:
        """Return the string representation of a bitset."""
        values = str(set(self)) if self else ""
        return f"{self.__class__.__name__}({values})"

    def add(self, bit: int) -> None:
        """Add `bit` to the set.

        Raises
        ------
        ValueError
            If `bit` is negative

        """
        if bit < 0:
            raise ValueError(f"bit not greater than or equal to 0, bit == {bit}")
        self.len += bit not in self
        self.bits |= 1 << bit

    def discard(self, bit: int) -> None:
        """Remove `bit` from the set.

        Raises
        ------
        ValueError
            If `bit` is negative

        """
        if bit < 0:
            raise ValueError(f"bit not greater than or equal to 0, bit == {bit}")
        self.len -= bit in self
        self.bits &= ~(1 << bit)



"""Module implementing an abstraction for navigation of array-backed trees."""

from typing import MutableSequence, Sequence, TypeVar

from .bitset import BitSet

T = TypeVar("T", covariant=True)


def reprtree(nodes: Sequence[T], *, fanout: int, indent: str = 4 * " ") -> str:
    """Return a string representation of `nodes`.

    Parameters
    ----------
    nodes
        A sequence of nodes in a tree.
    fanout
        The number of children per node.
    indent
        The prefix number of spaces to print at each level.

    """
    # track the current level of the tree and the index of the current node
    level_index_stack = [(0, 0)]

    # store the nodes that we've seen
    seen = BitSet()
    node_repr_pieces: MutableSequence[str] = []

    while level_index_stack:
        level, node_index = level_index_stack.pop()
        if node_index not in seen:
            node = nodes[node_index]
            node_repr_pieces.append(f"{level * indent}|-- {node}")
            node_indices = (
                fanout * node_index + i + 1 for i in reversed(range(fanout))
            )
            level_index_stack.extend(
                (level + 1, index) for index in node_indices if index < len(nodes)
            )
            seen.add(node_index)
    return "\n".join(node_repr_pieces)


def first_node(level: int, *, fanout: int) -> int:
    """Return the first node at `level`."""
    return (fanout ** level - 1) // (fanout - 1)


def last_node(level: int, *, fanout: int) -> int:
    """Return the last node at `level`."""
    return (fanout ** (level + 1) - 1) // (fanout - 1)


class IndexTree:
    """Abstraction for navigating around array-backed trees."""

    __slots__ = "height", "fanout", "nodes"

    def __init__(self, *, height: int, fanout: int) -> None:
        """Construct an :class:`~stupidb.associative.indextree.IndexTree`."""
        self.height = height
        self.fanout = fanout
        self.nodes = range((fanout ** height - 1) // (fanout - 1))

    @property
    def leaves(self) -> range:
        """Return the indices of the leaves of the tree."""
        height = self.height - 1
        first = self.first_node(height)
        last = self.last_node(height)
        return self.nodes[first:last]

    def __repr__(self) -> str:
        return reprtree(self.nodes, fanout=self.fanout)

    def __len__(self) -> int:
        """Return the number of nodes in the tree."""
        return len(self.nodes)

    def first_node(self, level: int) -> int:
        """Return the first node at `level`."""
        return first_node(level, fanout=self.fanout)

    def last_node(self, level: int) -> int:
        """Return the last node at `level`."""
        return last_node(level, fanout=self.fanout)

    def parent(self, node: int) -> int:
        """Return the index of the parent node of `node`."""
        if not node:
            parent_node_index = 0
        else:
            parent_node_index = (node - 1) // self.fanout

        # parent should never be negative
        assert (
            parent_node_index >= 0
        ), f"parent_node_index < 0: parent_node_index == {parent_node_index}"
        return parent_node_index



"""StupiDB. The stupidest database you'll ever come across.

This is project designed to illustate the concepts that underly a typical
relational database implementation, starting at naive execution of table-stakes
features up to rule-based query optimization.

.. warning::
   Please do not use this for any other reason than learning. There are no
   guarantees here except that there will be bugs.

"""

from __future__ import annotations

import abc
import collections
import functools
import itertools
import typing
from typing import Any, Generic, Iterable, Iterator, Mapping

import toolz

from .aggregation import (
    AggregateSpecification,
    Nulls,
    WindowAggregateSpecification,
    row_key_compare,
)
from .functions.associative.core import AssociativeAggregate
from .row import AbstractRow, JoinedRow, Row
from .typehints import (
    JoinPredicate,
    OrderBy,
    PartitionBy,
    PartitionKey,
    Predicate,
    Projector,
)


class Relation(abc.ABC):
    """An abstract relation."""

    __slots__ = ("partitioners",)

    def __init__(self) -> None:
        self.partitioners: Mapping[str, PartitionBy] = {}

    def __iter__(self) -> Iterator[AbstractRow]:
        """Iterate over the rows of a :class:`~stupidb.stupidb.Relation`.

        This method will reify rows with a new row identifier equal to the row number.

        """
        return (
            Row.from_mapping(row, _id=id)
            for id, row in enumerate(filter(None, self._produce()))
        )

    @abc.abstractmethod
    def _produce(self) -> Iterator[AbstractRow]:
        """Iterate over the rows of a :class:`~stupidb.stupidb.Relation`.

        Specific relation should implement this without reifying the row with
        the row identifier if possible. Reification is handled in the
        :meth:`~stupidb.stupidb.Relation.__iter__` method.

        """

    def __repr__(self) -> str:
        from stupidb.api import pretty

        return pretty(self)


class Table(Relation):
    __slots__ = ("rows",)

    def __init__(self, rows: Iterable[AbstractRow]) -> None:
        super().__init__()
        self.rows = rows

    @classmethod
    def from_iterable(cls, iterable: Iterable[Mapping[str, Any]]) -> Table:
        return cls(
            Row.from_mapping(mapping, _id=i) for i, mapping in enumerate(iterable)
        )

    def _produce(self) -> Iterator[AbstractRow]:
        return iter(self.rows)


class Projection(Relation):
    """A relation representing column selection.

    Attributes
    ----------
    child
    aggregations
    projections

    """

    __slots__ = "child", "aggregations", "projections"

    def __init__(
        self,
        child: Relation,
        projections: Mapping[str, Projector | WindowAggregateSpecification],
    ) -> None:
        super().__init__()
        self.child = child
        self.aggregations: Mapping[str, WindowAggregateSpecification] = {
            aggname: aggspec
            for aggname, aggspec in projections.items()
            if isinstance(aggspec, WindowAggregateSpecification)
        }
        self.projections: Mapping[str, Projector] = {
            name: projector
            for name, projector in projections.items()
            if callable(projector)
        }

    def _produce(self) -> Iterator[AbstractRow]:
        aggregations = self.aggregations
        # we need a row iterator for every aggregation to be fully generic
        # since they potentially share no structure
        #
        # one child iter for *all* projections
        # one child iter for *each* window aggregation
        child, *rowterators = itertools.tee(self.child, len(aggregations) + 1)
        aggnames = aggregations.keys()
        aggvalues = aggregations.values()

        # The .compute method returns an iterator of aggregation results
        # Each element of the iterator is the result of a single column in a
        # single row of the corresponding window function
        aggrows = (
            dict(zip(aggnames, aggrow))
            for aggrow in zip(
                *map(
                    WindowAggregateSpecification.compute,
                    aggvalues,
                    rowterators,
                )
            )
        )

        projections = self.projections
        projnames = projections.keys()
        projvalues = projections.values()
        projrows = (
            dict(zip(projnames, (proj(row) for proj in projvalues))) for row in child
        )

        # Use zip_longest here, because either of aggrows or projrows can be
        # empty
        return (
            Row(toolz.merge(projrow, aggrow), _id=-1)
            for aggrow, projrow in itertools.zip_longest(
                aggrows, projrows, fillvalue={}
            )
        )


class Mutate(Projection):
    """A relation representing appending columns to an existing relation."""

    __slots__ = ()

    def _produce(self) -> Iterator[AbstractRow]:
        # reasign self.child here to avoid clobbering its iteration
        # we need to use it twice: once for the computed columns (self.child)
        # used during the iteration of super().__iter__() and once for the
        # original relation (child)
        child, self.child = itertools.tee(self.child)
        return (
            Row.from_mapping(row, _id=-1)
            for row in map(toolz.merge, child, super()._produce())
        )


class Aggregation(Generic[AssociativeAggregate], Relation):
    """A relation representing aggregation of columns."""

    __slots__ = "child", "metrics"

    def __init__(
        self,
        child: Relation,
        metrics: Mapping[str, AggregateSpecification[AssociativeAggregate]],
    ) -> None:
        super().__init__()
        self.child = child
        self.metrics: Mapping[
            str, AggregateSpecification[AssociativeAggregate]
        ] = metrics

    def _produce(self) -> Iterator[AbstractRow]:
        aggregations = self.metrics

        # initialize aggregates
        grouped_aggs: Mapping[
            PartitionKey, Mapping[str, AssociativeAggregate]
        ] = collections.defaultdict(
            lambda: {
                name: aggspec.aggregate_type() for name, aggspec in aggregations.items()
            }
        )

        child = typing.cast(Relation, self.child)
        for row in child:
            key = tuple(
                (name, keyfunc(row)) for name, keyfunc in child.partitioners.items()
            )
            for name, agg in grouped_aggs[key].items():
                inputs = (getter(row) for getter in aggregations[name].getters)
                agg.step(*inputs)

        for grouping_key, aggs in grouped_aggs.items():
            data = dict(grouping_key)
            data.update((name, agg.finalize()) for name, agg in aggs.items())
            yield Row.from_mapping(data)


class Selection(Relation):
    """A relation of rows selected based on a predicate.

    Attributes
    ----------
    predicate
        A callable that takes an :class:`~stupidb.row.AbstractRow` and returns
        a :class:`bool`.

    """

    __slots__ = "child", "predicate"

    def __init__(self, child: Relation, predicate: Predicate) -> None:
        super().__init__()
        self.child = child
        self.predicate = predicate

    def _produce(self) -> Iterator[AbstractRow]:
        return filter(self.predicate, self.child)


class GroupBy(Relation):
    """A relation representing a partitioning of rows by a key.

    Attributes
    ----------
    group_by
        A callable that takes an :class:`~stupidb.row.AbstractRow` and returns
        an instance of :class:`typing.Hashable`.

    """

    __slots__ = "child", "group_by", "partitioners"

    def __init__(self, child: Relation, group_by: Mapping[str, PartitionBy]) -> None:
        super().__init__()
        self.child = child
        self.partitioners = group_by

    def _produce(self) -> Iterator[AbstractRow]:
        return iter(self.child)


class SortBy(Relation):
    """A relation representing rows of its child sorted by one or more keys.

    Attributes
    ----------
    order_by
        A callable that takes an :class:`~stupidb.row.AbstractRow` and returns
        an instance of :class:`~stupidb.protocols.Comparable`.
    null_ordering
        Whether to place the nulls of a column first or last.

    """

    __slots__ = "child", "order_by", "null_ordering"

    def __init__(
        self, child: Relation, order_by: tuple[OrderBy, ...], null_ordering: Nulls
    ) -> None:
        super().__init__()
        self.child = child
        self.order_by = order_by
        self.null_ordering = null_ordering

    def _produce(self) -> Iterator[AbstractRow]:
        return iter(
            sorted(
                self.child,
                key=functools.cmp_to_key(
                    functools.partial(
                        row_key_compare,
                        toolz.juxt(*self.order_by),
                        self.null_ordering,
                    )
                ),
            )
        )


class Limit(Relation):
    __slots__ = "child", "offset", "limit"

    def __init__(self, child: Relation, *, offset: int, limit: int | None) -> None:
        super().__init__()
        self.child = child
        self.offset = offset
        self.limit = limit

    def _produce(self) -> Iterator[AbstractRow]:
        limit = self.limit
        offset = self.offset
        return itertools.islice(
            self.child,
            offset,
            None if limit is None else offset + limit,
        )


class Join(Relation):
    __slots__ = "grouped", "rows"

    def __init__(self, left: Relation, right: Relation) -> None:
        super().__init__()
        self.grouped = itertools.groupby(
            (
                JoinedRow(left_row, right_row, _id=-1)
                for left_row, right_row in itertools.product(left, right)
            ),
            key=lambda row: row.left,
        )
        self.rows = itertools.chain.from_iterable(rows for _, rows in self.grouped)


class CrossJoin(Join):
    __slots__ = ()

    def _produce(self) -> Iterator[AbstractRow]:
        return iter(self.rows)


class InnerJoin(Join):
    __slots__ = ("predicate",)

    def __init__(
        self, left: Relation, right: Relation, predicate: JoinPredicate
    ) -> None:
        super().__init__(left, right)
        self.predicate = predicate

    def _produce(self) -> Iterator[AbstractRow]:
        return (row for row in self.rows if self.predicate(row.left, row.right))


class LeftJoin(Join):
    __slots__ = ("predicate",)

    def __init__(
        self, left: Relation, right: Relation, predicate: JoinPredicate
    ) -> None:
        super().__init__(left, right)
        self.predicate = predicate

    def _produce(self) -> Iterator[AbstractRow]:
        for left_row, joined_rows in self.grouped:
            matched = False

            for joined_row in joined_rows:
                right_row = joined_row.right
                if self.predicate(left_row, right_row):
                    matched = True
                    yield JoinedRow(left_row, right_row, _id=-1)
            if not matched:
                yield JoinedRow(left_row, dict.fromkeys(right_row), _id=-1)


class RightJoin(LeftJoin):
    __slots__ = ()

    def __init__(
        self, left: Relation, right: Relation, predicate: JoinPredicate
    ) -> None:
        super().__init__(right, left, predicate)

    def _produce(self) -> Iterator[AbstractRow]:
        for row in super()._produce():
            yield JoinedRow(row.right, row.left)


class SetOperation(Relation):
    """An abstract set operation."""

    __slots__ = "left", "right"

    def __init__(self, left: Relation, right: Relation) -> None:
        super().__init__()
        self.left = left
        self.right = right

    @staticmethod
    def itemize(
        mappings: Iterable[AbstractRow],
    ) -> frozenset[tuple[tuple[str, Any], ...]]:
        """Return a hashable version of `mappings`."""
        return frozenset(tuple(mapping.items()) for mapping in mappings)


class Union(SetOperation):
    """Union between two relations."""

    __slots__ = ()

    def _produce(self) -> Iterator[AbstractRow]:
        return toolz.unique(
            itertools.chain(self.left, self.right),
            key=lambda row: frozenset(row.items()),
        )


class UnionAll(SetOperation):
    """Non-unique union between two relations."""

    __slots__ = ()

    def _produce(self) -> Iterator[AbstractRow]:
        return itertools.chain(self.left, self.right)


class IntersectAll(SetOperation):
    """Non-unique intersection between two relations."""

    __slots__ = ()

    def _produce(self) -> Iterator[AbstractRow]:
        left_set = self.itemize(self.left)
        right_set = self.itemize(self.right)
        left_filtered = (row_items for row_items in left_set if row_items in right_set)
        right_filtered = (row_items for row_items in right_set if row_items in left_set)
        return (
            Row.from_mapping(dict(row))
            for row in itertools.chain(left_filtered, right_filtered)
        )


class Intersect(SetOperation):
    """Intersection of two relations."""

    __slots__ = ()

    def _produce(self) -> Iterator[AbstractRow]:
        return (
            Row.from_mapping(dict(row))
            for row in self.itemize(self.left) & self.itemize(self.right)
        )


class Difference(SetOperation):
    """Unique difference between two relations."""

    __slots__ = ()

    def _produce(self) -> Iterator[AbstractRow]:
        right_set = self.itemize(self.right)
        return toolz.unique(
            Row.from_mapping(dict(row_items))
            for row_items in (tuple(row.items()) for row in self.left)
            if row_items not in right_set
        )


class DifferenceAll(SetOperation):
    """Non-unique difference between two relations."""

    __slots__ = ()

    def _produce(self) -> Iterator[AbstractRow]:
        right_set = self.itemize(self.right)
        return (
            Row.from_mapping(dict(row_items))
            for row_items in (tuple(row.items()) for row in self.left)
            if row_items not in right_set
        )



"""Base aggregator interface."""

from __future__ import annotations

import abc
from typing import Generic, Sequence, TypeVar

from .row import AbstractRow
from .typehints import Getter, Output, Result, T

AggClass = TypeVar("AggClass", covariant=True)


class Aggregator(Generic[AggClass, Result], abc.ABC):
    """Interface for aggregators.

    Aggregators must implement the :meth:`~stupidb.aggregator.Aggregator.query`
    method. Aggregators are tied to a specific kind of aggregation. See the
    :meth:`~stupidb.aggregator.Aggregate.prepare` method for how to
    provide a custom aggregator.

    See Also
    --------
    stupidb.associative.segmenttree.SegmentTree
    stupidb.functions.navigation.core.NavigationAggregator

    """

    @abc.abstractmethod
    def __init__(self, arguments: Sequence[T], cls: AggClass) -> None:
        """Initialize an aggregator from `arguments` and `cls`."""

    @abc.abstractmethod
    def query(self, begin: int, end: int) -> Result | None:
        """Query the aggregator over the range from `begin` to `end`."""


class Aggregate(Generic[Output], abc.ABC):
    """An aggregate or window function."""

    __slots__ = ()

    @classmethod
    def prepare(
        cls,
        possible_peers: Sequence[AbstractRow],
        getters: tuple[Getter, ...],
        order_by_columns: Sequence[str],
    ) -> Aggregator[Aggregate[Output], Output]:
        """Prepare an aggregation of this type for computation."""
        arguments = [
            tuple(getter(peer) for getter in getters) for peer in possible_peers
        ]
        return cls.aggregator_class(arguments)

    @classmethod
    @abc.abstractmethod
    def aggregator_class(
        cls, inputs: Sequence[tuple[T | None, ...]]
    ) -> Aggregator[Aggregate[Output], Output]:  # pragma: no cover
        ...



"""Top-level package for stupidb."""

from stupidb.aggregation import Window  # noqa: F401
from stupidb.api import *  # noqa: F401,F403

try:
    import importlib.metadata as importlib_metadata
except ImportError:
    import importlib_metadata

__version__ = importlib_metadata.version(__name__)



from __future__ import annotations

import collections
import math
from typing import Generic, Iterator, MutableSequence, Sequence

from ..aggregator import Aggregator
from ..functions.associative.core import AssociativeAggregate
from ..typehints import Result, T
from . import indextree
from .bitset import BitSet


def make_segment_tree(
    leaf_arguments: Sequence[tuple[T, ...]],
    aggregate_type: type[AssociativeAggregate],
    *,
    fanout: int,
) -> Sequence[AssociativeAggregate]:
    """Make a segment tree from tuples `leaves` and class `aggregate_type`.

    The algorithm used here traverses from the bottom of tree upward, updating
    the parent every time a new node is seen.

    Parameters
    ----------
    leaves
        A sequence of tuples that make up the leaves of the segment tree
    aggregate_type
        The aggregate class whose instances compose the tree.

    """
    number_of_leaves = len(leaf_arguments)
    index_tree = indextree.IndexTree(
        height=int(math.ceil(math.log(number_of_leaves, fanout))) + 1,
        fanout=fanout,
    )
    num_nodes = len(index_tree)
    segment_tree_nodes: MutableSequence[AssociativeAggregate] = [
        aggregate_type() for _ in range(num_nodes)
    ]
    queue = collections.deque(index_tree.leaves)

    # seed the leaves
    for leaf_index, args in zip(queue, leaf_arguments):
        segment_tree_nodes[leaf_index].step(*args)

    seen = BitSet()

    while queue:
        node = queue.popleft()
        if node not in seen:
            seen.add(node)
            node_agg = segment_tree_nodes[node]
            parent = index_tree.parent(node)
            parent_agg = segment_tree_nodes[parent]
            parent_agg.combine(node_agg)
            if parent:
                # don't append the root, since we've already aggregated into
                # that node if parent == 0
                queue.append(parent)
    return segment_tree_nodes


class SegmentTree(
    Generic[T, AssociativeAggregate, Result],
    Aggregator[AssociativeAggregate, Result],
):
    """A segment tree for window aggregation.

    Attributes
    ----------
    nodes
        The nodes of the segment tree
    aggregate_type
        The class of the aggregate to use
    levels
        A list of the nodes in each level of the tree
    fanout
        The number of leaves to aggregate into each interior node

    """

    __slots__ = "nodes", "aggregate_type", "levels", "fanout"

    def __init__(
        self,
        leaves: Sequence[tuple[T | None, ...]],
        aggregate_type: type[AssociativeAggregate],
        *,
        fanout: int,
    ) -> None:
        """Construct a segment tree."""
        self.nodes: Sequence[AssociativeAggregate] = make_segment_tree(
            leaves, aggregate_type, fanout=fanout
        )
        self.aggregate_type: type[AssociativeAggregate] = aggregate_type
        self.fanout = fanout
        self.height = int(math.ceil(math.log(len(leaves), fanout))) + 1
        self.levels: Sequence[Sequence[AssociativeAggregate]] = list(
            self.iterlevels(self.nodes, fanout=fanout)
        )

    @staticmethod
    def iterlevels(
        nodes: Sequence[AssociativeAggregate], *, fanout: int
    ) -> Iterator[Sequence[AssociativeAggregate]]:
        """Iterate over every level in the tree.

        Parameters
        ----------
        nodes
            The nodes of the tree whose levels will be yielded.
        fanout
            The number child nodes per interior node

        """
        height = int(math.ceil(math.log(len(nodes), fanout)))
        for level in range(height):
            start = indextree.first_node(level, fanout=fanout)
            stop = indextree.last_node(level, fanout=fanout)
            yield nodes[start:stop]

    def __repr__(self) -> str:
        return indextree.reprtree(self.nodes, fanout=self.fanout)

    def query(self, begin: int, end: int) -> Result | None:
        """Aggregate the values between `begin` and `end` using `aggregate`.

        Parameters
        ----------
        begin
            The start of the range to aggregate
        end
            The end of the range to aggregate

        """
        fanout = self.fanout
        aggregate: AssociativeAggregate = self.aggregate_type()

        for level in reversed(self.levels):
            parent_begin = begin // fanout
            parent_end = end // fanout
            if parent_begin == parent_end:
                for item in level[begin:end]:
                    aggregate.combine(item)
                return aggregate.finalize()
            group_begin = parent_begin * fanout
            if begin != group_begin:
                limit = group_begin + fanout
                for item in level[begin:limit]:
                    aggregate.combine(item)
                parent_begin += 1
            group_end = parent_end * fanout
            if end != group_end:
                for item in level[group_end:end]:
                    aggregate.combine(item)
            begin = parent_begin
            end = parent_end
        return None  # pragma: no cover



"""Navigation and simple window function interface and implementation."""

from __future__ import annotations

import abc
from typing import Any, Sequence, Union

from ...protocols import Comparable
from ...typehints import T
from .core import RankingAggregate


class RowNumber(RankingAggregate[int]):
    """Row number analytic function."""

    __slots__ = ("row_number",)

    def __init__(
        self, order_by_values: Sequence[tuple[Comparable | None, ...]]
    ) -> None:
        super().__init__(order_by_values)
        self.row_number = 0

    def execute(self, begin: int, end: int) -> int:
        """Compute an abstract row rank value for rows between `begin` and `end`."""
        row_number = self.row_number
        self.row_number += 1
        return row_number


class Sentinel:
    """A class that is not equal to anything except instances of itself.

    This class is used as the starting value for :class:`stupidb.ranking.Rank`
    and :class:`stupidb.ranking.DenseRank` because their algorithms compare the
    previous ``ORDER BY`` value in the sequence to determine whether to
    increase the rank.

    """

    __slots__ = ()

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self))


Either = Union[Sentinel, T]


class AbstractRank(RowNumber):
    """A class represnting a numerical ordering of rows."""

    __slots__ = ("previous_value",)

    def __init__(
        self, order_by_values: Sequence[tuple[Comparable | None, ...]]
    ) -> None:
        super().__init__(order_by_values)
        self.previous_value: Either | None = Sentinel()

    @abc.abstractmethod
    def rank(self, current_order_by_value: Comparable, current_row_number: int) -> int:
        """Compute the rank of the current row."""

    def execute(self, begin: int, end: int) -> int:
        """Compute an abstract row rank value for rows between `begin` and `end`."""
        current_row_number = super().execute(begin, end)
        current_order_by_value = self.order_by_values[current_row_number]
        rank = self.rank(current_order_by_value, current_row_number)
        assert (
            rank >= 0
        ), f"rank should be greater than or equal to 0, got rank == {rank:d}"
        self.previous_value = current_order_by_value
        assert not isinstance(
            self.previous_value, Sentinel
        ), "expected non-Sentinel order by value, got Sentinel"
        return rank


class Rank(AbstractRank):
    """Non-dense ranking computation."""

    __slots__ = ("previous_rank",)

    def __init__(
        self, order_by_values: Sequence[tuple[Comparable | None, ...]]
    ) -> None:
        super().__init__(order_by_values)
        self.previous_rank = -1

    def rank(self, current_order_by_value: Comparable, current_row_number: int) -> int:
        """Rank the current row according to `current_order_by_value`."""
        if current_order_by_value != self.previous_value:
            rank = current_row_number
        else:
            rank = self.previous_rank
        self.previous_rank = rank
        return rank


class DenseRank(AbstractRank):
    """Dense ranking computation."""

    __slots__ = ("current_rank",)

    def __init__(
        self, order_by_values: Sequence[tuple[Comparable | None, ...]]
    ) -> None:
        super().__init__(order_by_values)
        self.current_rank = -1

    def rank(self, current_order_by_value: Comparable, current_row_number: int) -> int:
        """Compute the current rank, densely."""
        self.current_rank += current_order_by_value != self.previous_value
        return self.current_rank



"""Tests for `stupidb` package."""

from __future__ import annotations

import builtins
import itertools
import operator
import sqlite3
import statistics
from datetime import date, timedelta
from typing import Any, Callable, Iterable, Iterator, Mapping, Sequence, TypeVar

import pytest
import toolz

from stupidb.aggregation import Window
from stupidb.api import (
    aggregate,
    const,
    count,
    cov_pop,
    cov_samp,
    cross_join,
    exists,
    full_join,
    get,
    group_by,
    inner_join,
    left_join,
    limit,
    max,
    mean,
    min,
    mutate,
    nth,
    order_by,
    over,
    pretty,
    right_join,
    select,
    sift,
    stdev_pop,
    stdev_samp,
    sum,
    table,
    total,
    var_pop,
    var_samp,
)
from stupidb.core import Relation, Table
from stupidb.row import Row

from .conftest import Element, assert_rowset_equal


def test_projection(rows: list[dict[str, Element]]) -> None:
    expected: list[Mapping[str, Element]] = [
        dict(z="a", c=1, d=2),
        dict(z="b", c=2, d=-1),
        dict(z="a", c=3, d=4),
        dict(z="a", c=4, d=-3),
        dict(z="a", c=1, d=-3),
        dict(z="b", c=2, d=-3),
        dict(z="b", c=3, d=-3),
    ]
    pipeline = table(rows) >> select(c=get("a"), d=get("b"), z=get("z"))
    result = list(pipeline)
    assert_rowset_equal(result, expected)


def test_selection(rows: list[dict[str, Element]]) -> None:
    expected: list[Mapping[str, Element]] = [
        dict(z="a", c=1, d=2),
        dict(z="a", c=1, d=-3),
        dict(z="a", c=4, d=-3),
        dict(z="a", c=3, d=4),
        dict(z="b", c=2, d=-1),
        dict(z="b", c=2, d=-3),
        dict(z="b", c=3, d=-3),
    ]
    selection = (
        table(rows)
        >> select(c=lambda r: r["a"], d=lambda r: r["b"], z=lambda r: r["z"])
        >> sift(lambda r: True)
    )
    assert_rowset_equal(selection, expected)


def test_group_by(rows: list[dict[str, Element]]) -> None:
    expected: list[Mapping[str, Element]] = [
        {"c": 1, "mean": -0.5, "total": -1, "z": "a"},
        {"c": 2, "mean": -2.0, "total": -4, "z": "b"},
        {"c": 3, "mean": 4.0, "total": 4, "z": "a"},
        {"c": 4, "mean": -3.0, "total": -3, "z": "a"},
        {"c": 3, "mean": -3.0, "total": -3, "z": "b"},
    ]
    gb = (
        table(rows)
        >> select(c=lambda r: r["a"], d=lambda r: r["b"], z=lambda r: r["z"])
        >> sift(lambda _: True)
        >> group_by(c=lambda r: r["c"], z=lambda r: r["z"])
        >> aggregate(total=sum(lambda r: r["d"]), mean=mean(lambda r: r["d"]))
    )
    result = list(gb)
    assert_rowset_equal(result, expected)


def subclasses(cls: type) -> frozenset[type]:
    classes = cls.__subclasses__()
    return (
        frozenset({cls})
        | frozenset(classes)
        | frozenset(itertools.chain.from_iterable(map(subclasses, classes)))
    )


@pytest.mark.parametrize(  # type: ignore[misc]
    "cls",
    (cls for cls in subclasses(Relation) if cls is not Relation and cls is not Table),
)
def test_from_iterable_is_invalid(cls: Relation) -> None:
    with pytest.raises(AttributeError):
        cls.from_iterable([dict(a=1)])


def test_cross_join(
    left: list[dict[str, Element]], right: list[dict[str, Element]]
) -> None:
    join = (
        table(left)
        >> cross_join(table(right))
        >> select(left_z=lambda r: r.left["z"], right_z=lambda r: r.right["z"])
    )
    result = list(join)
    assert len(result) == len(left) * len(right)
    expected = [
        {"left_z": left["z"], "right_z": right["z"]}
        for left, right in list(itertools.product(left, right))
    ]
    assert len(expected) == len(result)
    assert_rowset_equal(result, expected)


def test_inner_join(
    left: list[dict[str, Element]],
    right: list[dict[str, Element]],
    con: sqlite3.Connection,
) -> None:
    join = (
        table(left)
        >> inner_join(
            table(right),
            lambda left, right: (
                left["z"] == "a" and right["z"] == "a" and left["a"] == right["a"]
            ),
        )
        >> select(
            left_a=lambda r: r.left["a"],
            right_a=lambda r: r.right["a"],
            right_z=lambda r: r.right["z"],
            left_z=lambda r: r.left["z"],
        )
    )
    result = list(join)
    query = """
SELECT
    t.a AS left_a,
    s.a AS right_a,
    s.z AS right_z,
    t.z AS left_z
FROM left t
INNER JOIN right s
ON t.z = 'a'
  AND s.z = 'a'
  AND t.a = s.a"""
    expected = [
        dict(left_a=left_a, right_a=right_a, right_z=right_z, left_z=left_z)
        for left_a, right_a, right_z, left_z in con.execute(query).fetchall()
    ]
    assert_rowset_equal(result, expected)


def test_left_join(
    left: list[dict[str, Element]],
    right: list[dict[str, Element]],
    con: sqlite3.Connection,
) -> None:
    join = (
        table(left)
        >> left_join(table(right), lambda left, right: left["z"] == right["z"])
        >> select(left_z=lambda r: r.left["z"], right_z=lambda r: r.right["z"])
    )
    result = list(join)
    query = """
SELECT t.z AS left_z, s.z AS right_z
FROM left t
LEFT OUTER JOIN right s
ON t.z = s.z"""
    expected = [
        dict(left_z=left_z, right_z=right_z)
        for left_z, right_z in con.execute(query).fetchall()
    ]
    assert_rowset_equal(result, expected)


def test_right_join(
    left: list[dict[str, Element]],
    right: list[dict[str, Element]],
    con: sqlite3.Connection,
) -> None:
    join = (
        table(left)
        >> right_join(table(right), lambda left, right: left["z"] == right["z"])
        >> select(left_z=lambda r: r.left["z"], right_z=lambda r: r.right["z"])
    )
    result = list(join)
    query = """
SELECT t.z AS left_z, s.z AS right_z
FROM right s
LEFT OUTER JOIN left t
ON t.z = s.z"""
    expected = [
        dict(left_z=left_z, right_z=right_z)
        for left_z, right_z in con.execute(query).fetchall()
    ]
    assert_rowset_equal(result, expected)


@pytest.mark.xfail(  # type: ignore[misc]
    raises=NotImplementedError,
    reason="full outer joins are not yet supported",
)
def test_full_join(
    employee: list[dict[str, Element]],
    department: list[dict[str, Element]],
    con: sqlite3.Connection,
) -> None:  # pragma: no cover
    join = (
        table(employee)
        >> full_join(
            table(department), lambda e, d: e["department_id"] == d["department_id"]
        )
        >> select(
            last_name=lambda row: row.left["last_name"],
            e_dep_id=lambda row: row.left["department_id"],
            department_name=lambda row: row.right["department_name"],
            d_dep_id=lambda row: row.right["department_id"],
        )
    )
    result = list(join)
    query = """
with
left_outer as (
    SELECT
        e.last_name,
        e.department_id as e_dep_id,
        d.department_name,
        d.department_id AS d_dep_id
    FROM employee e
    LEFT OUTER JOIN department d
    USING (department_id)
),
right_outer as (
    SELECT
        e.last_name,
        e.department_id as e_dep_id,
        d.department_name,
        d.department_id AS d_dep_id
    FROM department d
    LEFT OUTER JOIN employee e
    USING (department_id)
)
SELECT * FROM left_outer
UNION
SELECT * FROM right_outer
"""
    expected = [
        dict(zip(("last_name", "e_dep_id", "department_name", "d_dep_id"), row))
        for row in con.execute(query).fetchall()
    ]
    breakpoint()
    assert_rowset_equal(result, expected)


def test_semi_join() -> None:
    rows = [
        dict(z="a", a=1, b=2),
        dict(z="b", a=2, b=-1),
        dict(z="a", a=3, b=4),
        dict(z="a", a=4, b=-3),
        dict(z="a", a=1, b=-3),
        dict(z="b", a=2, b=-3),
        dict(z="b", a=3, b=-3),
    ]
    other_rows = [
        dict(z="a", a=4, b=-3),
        dict(z="a", a=1, b=-3),
        dict(z="b", a=2, b=-3),
        dict(z="b", a=3, b=-3),
    ]

    pipeline = table(rows) >> sift(
        lambda r: exists(table(other_rows) >> sift(lambda o: r["z"] == o["z"]))
    )
    result = list(pipeline)
    assert result == rows


def test_semi_join_not_all_rows_match() -> None:
    rows = [
        dict(z="a", a=1, b=2),
        dict(z="b", a=2, b=-1),
        dict(z="a", a=3, b=4),
        dict(z="a", a=4, b=-3),
        dict(z="a", a=1, b=-3),
        dict(z="b", a=2, b=-3),
        dict(z="b", a=3, b=-3),
    ]
    other_rows = [dict(z="b", a=2, b=-3), dict(z="b", a=3, b=-3)]

    pipeline = table(rows) >> sift(
        lambda r: exists(table(other_rows) >> sift(lambda o: r["z"] == o["z"]))
    )
    result = list(pipeline)
    expected = [row for row in rows if row["z"] == "b"]
    assert result == expected


@pytest.mark.parametrize("offset", range(4))  # type: ignore[misc]
@pytest.mark.parametrize("lim", range(4))  # type: ignore[misc]
def test_valid_limit(rows: Sequence[Mapping[str, Any]], offset: int, lim: int) -> None:
    pipeline = table(rows) >> limit(lim, offset=offset)
    assert list(pipeline) == rows[offset : offset + lim]


@pytest.mark.parametrize(  # type: ignore[misc]
    ("offset", "lim"),
    ((offset, lim) for lim in range(-2, 1) for offset in range(-2, 1) if offset or lim),
)
def test_invalid_limit(
    rows: Sequence[Mapping[str, Any]],
    offset: int,
    lim: int,
) -> None:
    with pytest.raises(ValueError):
        table(rows) >> limit(lim, offset=offset)


def test_right_shiftable(rows: list[dict[str, Element]]) -> None:
    pipeline = (
        table(rows)
        >> select(c=lambda r: r["a"], d=lambda r: r["b"], z=lambda r: r["z"])
        >> sift(lambda r: True)
        >> group_by(c=lambda r: r["c"], z=lambda r: r["z"])
        >> aggregate(
            total=sum(lambda r: r["d"]),
            mean=mean(lambda r: r["d"]),
            my_samp_cov=cov_samp(lambda r: r["d"], lambda r: r["d"]),
            my_pop_cov=cov_pop(lambda r: r["d"], lambda r: r["d"]),
        )
    )

    expected: list[Mapping[str, Element]] = [
        {
            "c": 1,
            "mean": -0.5,
            "my_samp_cov": 15.0,
            "my_pop_cov": 7.5,
            "total": -1,
            "z": "a",
        },
        {
            "c": 2,
            "mean": -2.0,
            "my_samp_cov": 3.0,
            "my_pop_cov": 1.5,
            "total": -4,
            "z": "b",
        },
        {
            "c": 3,
            "mean": 4.0,
            "my_samp_cov": None,
            "my_pop_cov": 0.0,
            "total": 4,
            "z": "a",
        },
        {
            "c": 4,
            "mean": -3.0,
            "my_samp_cov": None,
            "my_pop_cov": 0.0,
            "total": -3,
            "z": "a",
        },
        {
            "c": 3,
            "mean": -3.0,
            "my_samp_cov": None,
            "my_pop_cov": 0.0,
            "total": -3,
            "z": "b",
        },
    ]
    result = list(pipeline)
    assert_rowset_equal(result, expected)


def test_rows_window(rows: list[dict[str, Element]]) -> None:
    pipeline = (
        table(rows)
        >> mutate(
            my_agg=sum(lambda r: r["a"])
            >> over(
                Window.rows(
                    order_by=[lambda r: r["e"]],
                    partition_by=[lambda r: r["z"]],
                    preceding=lambda r: 2,
                    following=lambda r: 0,
                )
            )
        )
        >> order_by(lambda r: r["z"], lambda r: r["e"])
    )
    result = list(pipeline)
    expected_aggrows = [
        {"my_agg": 1},
        {"my_agg": 2},
        {"my_agg": 4},
        {"my_agg": 8},
        {"my_agg": 8},
        {"my_agg": 4},
        {"my_agg": 7},
    ]
    expected = sorted(
        map(toolz.merge, rows, expected_aggrows),
        key=lambda r: (r["z"], r["e"]),
    )
    assert len(result) == len(expected)
    assert_rowset_equal(result, expected)


def test_rows_window_partition(rows: list[dict[str, Element]]) -> None:
    pipeline = (
        table(rows)
        >> mutate(
            my_agg=sum(lambda r: r.a) >> over(Window.rows(partition_by=[lambda r: r.z]))
        )
        >> order_by(lambda r: r.z, lambda r: r.e)
    )
    result = list(pipeline)
    expected_aggrows = [
        {"my_agg": 9},
        {"my_agg": 7},
        {"my_agg": 9},
        {"my_agg": 9},
        {"my_agg": 9},
        {"my_agg": 7},
        {"my_agg": 7},
    ]
    assert len(result) == len(rows)
    assert len(result) == len(expected_aggrows)
    expected = sorted(
        map(toolz.merge, rows, expected_aggrows),
        key=lambda r: (r["z"], r["e"]),
    )
    assert len(result) == len(expected)
    assert_rowset_equal(result, expected)


def test_range_window_with_multiple_ordering_keys_fails(
    rows: list[dict[str, Element]]
) -> None:
    with pytest.raises(ValueError):
        table(rows) >> mutate(
            my_agg=sum(lambda r: r["a"])
            >> over(
                Window.range(
                    order_by=[lambda r: r["e"], lambda r: r["a"]],
                    partition_by=[lambda r: r["z"]],
                    preceding=lambda r: 2,
                    following=lambda r: 0,
                )
            )
        )


def test_range_window(rows: list[dict[str, Element]]) -> None:
    pipeline = (
        table(rows)
        >> mutate(
            my_agg=sum(lambda r: r["a"])
            >> over(
                Window.range(
                    order_by=[lambda r: r["e"]],
                    partition_by=[lambda r: r["z"]],
                    preceding=lambda r: 2,
                    following=lambda r: 0,
                )
            )
        )
        >> order_by(lambda r: r["z"], lambda r: r["e"])
        >> select(
            a=lambda r: r.a,
            e=lambda r: r.e,
            my_agg=lambda r: r.my_agg,
            z=lambda r: r.z,
        )
    )
    result = list(pipeline)
    expected_a: list[Mapping[str, Element]] = [
        {"a": 1, "e": 1, "my_agg": 1, "z": "a"},
        {"a": 3, "e": 3, "my_agg": 4, "z": "a"},
        {"a": 4, "e": 4, "my_agg": 7, "z": "a"},
        {"a": 1, "e": 5, "my_agg": 8, "z": "a"},
    ]
    expected_b: list[Mapping[str, Element]] = [
        {"a": 2, "e": 2, "my_agg": 2, "z": "b"},
        {"a": 2, "e": 6, "my_agg": 2, "z": "b"},
        {"a": 3, "e": 7, "my_agg": 5, "z": "b"},
    ]

    expected = expected_a + expected_b

    assert len(result) == len(expected)
    assert_rowset_equal(result, expected)


def test_temporal_range_window(
    t_table: Table, t_rows: list[dict[str, Element]]
) -> None:
    query = t_table >> mutate(
        avg_balance=mean(lambda r: r["balance"])
        >> over(
            Window.range(
                order_by=[lambda r: r["date"]],
                partition_by=[lambda r: r["name"]],
                preceding=lambda r: timedelta(days=3),
                following=lambda r: timedelta(days=0),
            )
        )
    )
    result = list(query)
    expected: list[Mapping[str, Element]] = [
        {
            "avg_balance": 2.0,
            "balance": 2,
            "date": date(2018, 1, 1),
            "name": "alice",
        },
        {
            "avg_balance": 3.0,
            "balance": 4,
            "date": date(2018, 1, 4),
            "name": "alice",
        },
        {
            "avg_balance": 0.5,
            "balance": -3,
            "date": date(2018, 1, 6),
            "name": "alice",
        },
        {
            "avg_balance": -0.666_666_666_666_666_6,
            "balance": -3,
            "date": date(2018, 1, 7),
            "name": "alice",
        },
        {
            "avg_balance": -1.0,
            "balance": -1,
            "date": date(2018, 1, 2),
            "name": "bob",
        },
        {
            "avg_balance": -2.0,
            "balance": -3,
            "date": date(2018, 1, 3),
            "name": "bob",
        },
        {
            "avg_balance": -2.333_333_333_333_333_5,
            "balance": -3,
            "date": date(2018, 1, 4),
            "name": "bob",
        },
    ]
    assert_rowset_equal(result, expected)


def test_agg(rows: list[dict[str, Element]]) -> None:
    pipeline = table(rows) >> aggregate(
        sum=sum(lambda r: r["e"]),
        mean=mean(lambda r: r["e"]),
        count=count(lambda r: r["e"]),
    )
    (result,) = list(pipeline)
    assert result["sum"] == 28
    assert result["mean"] == result["sum"] / result["count"]


def test_invalid_agg(rows: list[dict[str, Element]]) -> None:
    with pytest.raises(TypeError, match="Invalid projection"):
        select(
            not_an_agg=lambda r: r["e"],
            my_agg=sum(lambda r: r["e"]),
            my_agg2=mean(lambda r: r["e"]),
            my_count=count(lambda r: r["e"]),
        )


T = TypeVar("T")
U = TypeVar("U")


def cumagg(seq: Iterable[T], combine: Callable[[T, T], T]) -> Iterator[T]:
    """Cumulative aggregation."""
    it = iter(seq)
    result = next(it)
    yield result
    for value in it:
        result = combine(value, result)
        yield result


def test_cumagg(rows: list[dict[str, Element]]) -> None:
    window = Window.rows(order_by=[get("e")])
    query = table(rows) >> select(
        cumsum=sum(get("e")) >> over(window),
        cumcount=count(get("e")) >> over(window),
    )
    result = list(query)
    expected = [
        dict(cumsum=cs, cumcount=cc)
        for cs, cc in zip(cumagg(range(1, 8), operator.add), range(1, 8))
    ]
    assert_rowset_equal(result, expected)


def test_total_vs_sum() -> None:
    rows = [dict(value=None), dict(value=None)]
    query = table(rows) >> aggregate(
        sum=sum(lambda r: r.value), total=total(lambda r: r.value)
    )
    (result_row,) = list(query)
    assert result_row.sum is None
    assert result_row.total == 0


def test_minmax(rows: list[dict[str, Element]]) -> None:
    query = table(rows) >> aggregate(min=min(lambda r: r.e), max=max(lambda r: r.e))
    result = list(query)
    e = list(map(get("e"), rows))
    expected = [dict(min=builtins.min(e), max=builtins.max(e))]
    assert_rowset_equal(result, expected)


def test_min_max_window(t_rows: list[dict[str, Element]]) -> None:
    window = Window.range(partition_by=[lambda r: r.name])
    query = table(t_rows) >> select(
        min_date=min(lambda r: r.date) >> over(window),
        max_date=max(lambda r: r.date) >> over(window),
    )
    result = list(query)
    expected = [
        dict(min_date=date(2018, 1, 1), max_date=date(2018, 1, 7)),
        dict(min_date=date(2018, 1, 1), max_date=date(2018, 1, 7)),
        dict(min_date=date(2018, 1, 1), max_date=date(2018, 1, 7)),
        dict(min_date=date(2018, 1, 1), max_date=date(2018, 1, 7)),
        dict(min_date=date(2018, 1, 2), max_date=date(2018, 1, 4)),
        dict(min_date=date(2018, 1, 2), max_date=date(2018, 1, 4)),
        dict(min_date=date(2018, 1, 2), max_date=date(2018, 1, 4)),
    ]
    assert_rowset_equal(result, expected)


def test_variance_window(t_rows: list[dict[str, Element]]) -> None:
    window = Window.range(partition_by=[get("name")])
    query = table(t_rows) >> select(
        name=get("name"),
        var=var_samp(get("balance")) >> over(window),
        std=stdev_samp(get("balance")) >> over(window),
        popvar=var_pop(get("balance")) >> over(window),
        popstd=stdev_pop(get("balance")) >> over(window),
    )
    result = list(query)
    alice = [row["balance"] for row in t_rows if row["name"] == "alice"]
    bob = [row["balance"] for row in t_rows if row["name"] == "bob"]
    expected = {
        Row(
            dict(
                name="alice",
                var=statistics.variance(alice),
                std=statistics.stdev(alice),
                popvar=statistics.pvariance(alice),
                popstd=statistics.pstdev(alice),
            ),
        ),
        Row(
            dict(
                name="bob",
                var=statistics.variance(bob),
                std=statistics.stdev(bob),
                popvar=statistics.pvariance(bob),
                popstd=statistics.pstdev(bob),
            ),
        ),
    }
    assert set(result) == expected


def test_pretty(t_rows):
    window = Window.range(partition_by=[lambda r: r.name])
    query = table(t_rows) >> select(
        min_date=min(lambda r: r.date) >> over(window),
        max_date=max(lambda r: r.date) >> over(window),
    )
    expected = """\
min_date    max_date
----------  ----------
2018-01-01  2018-01-07
2018-01-01  2018-01-07
2018-01-01  2018-01-07
2018-01-01  2018-01-07
2018-01-02  2018-01-04
2018-01-02  2018-01-04
2018-01-02  2018-01-04"""
    result = query >> pretty()
    assert result == expected


def test_pretty_fmt(t_rows):
    window = Window.range(partition_by=[lambda r: r.name])
    query = table(t_rows) >> select(
        min_date=min(lambda r: r.date) >> over(window),
        max_date=max(lambda r: r.date) >> over(window),
    )
    expected = """\
╒════════════╤════════════╕
│ min_date   │ max_date   │
╞════════════╪════════════╡
│ 2018-01-01 │ 2018-01-07 │
├────────────┼────────────┤
│ 2018-01-01 │ 2018-01-07 │
├────────────┼────────────┤
│ 2018-01-01 │ 2018-01-07 │
├────────────┼────────────┤
│ 2018-01-01 │ 2018-01-07 │
├────────────┼────────────┤
│ 2018-01-02 │ 2018-01-04 │
├────────────┼────────────┤
│ 2018-01-02 │ 2018-01-04 │
├────────────┼────────────┤
│ 2018-01-02 │ 2018-01-04 │
╘════════════╧════════════╛"""
    result = query >> pretty(tablefmt="fancy_grid")
    assert result == expected


def test_multiple_windows(t_rows: list[dict[str, Element]]) -> None:
    query = table(t_rows) >> select(
        nth_date=(
            nth(get("date"), const(1)) >> over(Window.range(partition_by=[get("name")]))
        ),
        max_balance=(
            max(get("balance"))
            >> over(Window.range(partition_by=[lambda r: r.balance > 0]))
        ),
    )
    result = list(query)
    expected = [
        dict(nth_date=date(2018, 1, 4), max_balance=4),
        dict(nth_date=date(2018, 1, 4), max_balance=4),
        dict(nth_date=date(2018, 1, 4), max_balance=-1),
        dict(nth_date=date(2018, 1, 4), max_balance=-1),
        dict(nth_date=date(2018, 1, 3), max_balance=-1),
        dict(nth_date=date(2018, 1, 3), max_balance=-1),
        dict(nth_date=date(2018, 1, 3), max_balance=-1),
    ]
    assert_rowset_equal(result, expected)



from stupidb.associative.segmenttree import SegmentTree
from stupidb.functions.associative import Count, Mean, Min, SampleCovariance, Sum, Total


def test_repr_segment_tree_fanout_2() -> None:
    tree = SegmentTree([(1,), (2,), (3,)], aggregate_type=Sum, fanout=2)
    result = repr(tree)
    expected = """\
|-- Sum(total=6, count=3)
    |-- Sum(total=3, count=2)
        |-- Sum(total=1, count=1)
        |-- Sum(total=2, count=1)
    |-- Sum(total=3, count=1)
        |-- Sum(total=3, count=1)
        |-- Sum(total=None, count=0)"""
    assert result == expected


def test_repr_segment_tree_fanout_4() -> None:
    tree = SegmentTree([(1,), (2,), (3,)], aggregate_type=Sum, fanout=4)
    result = repr(tree)
    expected = """\
|-- Sum(total=6, count=3)
    |-- Sum(total=1, count=1)
    |-- Sum(total=2, count=1)
    |-- Sum(total=3, count=1)
    |-- Sum(total=None, count=0)"""
    assert result == expected


def test_count_repr() -> None:
    count = Count[str]()
    assert repr(count) == "Count(count=0)"
    count.step("a")
    assert repr(count) == "Count(count=1)"


def test_sum_repr() -> None:
    sum = Sum[int, int]()
    assert repr(sum) == "Sum(total=None, count=0)"
    sum.step(1)
    assert repr(sum) == "Sum(total=1, count=1)"


def test_total_repr() -> None:
    total = Total[int, int]()
    assert repr(total) == "Total(total=0, count=0)"
    total.step(1)
    assert repr(total) == "Total(total=1, count=1)"


def test_mean_repr() -> None:
    mean = Mean[int, float]()
    assert repr(mean) == "Mean(total=None, count=0, mean=None)"
    mean.step(3)
    assert repr(mean) == "Mean(total=3, count=1, mean=3.0)"
    mean.step(4)
    assert repr(mean) == "Mean(total=7, count=2, mean=3.5)"


def test_min_repr() -> None:
    min = Min()
    assert repr(min) == "Min(current_value=None)"
    min.step(2)
    assert repr(min) == "Min(current_value=2)"
    min.step(1)
    assert repr(min) == "Min(current_value=1)"
    min.step(3)
    assert repr(min) == "Min(current_value=1)"


def test_sample_covariance_repr() -> None:
    cov = SampleCovariance[float, float]()
    assert repr(cov) == "SampleCovariance(mean_x=0.0, mean_y=0.0, cov=0.0, count=0)"
    cov.step(1.0, 2.0)
    assert repr(cov) == "SampleCovariance(mean_x=2.0, mean_y=2.0, cov=0.0, count=1)"
    cov.step(3.0, 4.5)
    assert repr(cov) == "SampleCovariance(mean_x=5.0, mean_y=3.25, cov=1.25, count=2)"



from __future__ import annotations

import abc
from typing import Generic, Sequence, TypeVar

from ...aggregator import Aggregate, Aggregator
from ...typehints import Input1, Input2, Output, T

UA = TypeVar("UA", bound="UnaryAssociativeAggregate")
BA = TypeVar("BA", bound="BinaryAssociativeAggregate")


class AbstractAssociativeAggregate(Aggregate[Output]):
    """Base class for aggregations with an associative binary operation."""

    __slots__ = ("count",)

    def __init__(self) -> None:
        """Construct an abstract associative aggregate."""
        self.count = 0

    @abc.abstractmethod
    def finalize(self) -> Output | None:
        """Compute the value of the aggregation from its current state."""

    @classmethod
    def aggregator_class(
        cls, values: Sequence[tuple[T | None, ...]]
    ) -> Aggregator[Aggregate[Output], Output]:
        from ...associative.segmenttree import SegmentTree

        return SegmentTree(values, cls, fanout=4)


class UnaryAssociativeAggregate(
    AbstractAssociativeAggregate[Output],
    Generic[Input1, Output],
):
    """A an abstract associative aggregate that takes one argument."""

    __slots__ = ()

    @abc.abstractmethod
    def step(self, input1: Input1 | None) -> None:
        """Perform a single step of the aggregation."""

    @abc.abstractmethod
    def combine(self: UA, other: UA) -> None:
        """Combine two UnaryAssociativeAggregate instances."""

    @abc.abstractmethod
    def finalize(self) -> Output | None:
        """Compute the value of the aggregation from its current state."""


class BinaryAssociativeAggregate(
    AbstractAssociativeAggregate[Output],
    Generic[Input1, Input2, Output],
):
    """A an abstract associative aggregate that takes two arguments."""

    __slots__ = ()

    @abc.abstractmethod
    def step(self, input1: Input1 | None, input2: Input2 | None) -> None:
        """Perform a single step of the aggregation."""

    @abc.abstractmethod
    def combine(self: BA, other: BA) -> None:
        """Combine two BinaryAssociativeAggregate instances."""

    @abc.abstractmethod
    def finalize(self) -> Output | None:
        """Compute the value of the aggregation from its current state."""


AssociativeAggregate = TypeVar(
    "AssociativeAggregate",
    UnaryAssociativeAggregate,
    BinaryAssociativeAggregate,
)



from __future__ import annotations

from datetime import date
from typing import Mapping

from stupidb import Window, const, first, get, lag, last, lead, nth, over, select, table

from .conftest import Element, assert_rowset_equal


def test_first_last(t_rows: list[dict[str, Element]]) -> None:
    window = Window.range(partition_by=[get("name")])
    query = table(t_rows) >> select(
        first_date=first(get("date")) >> over(window),
        last_date=last(get("date")) >> over(window),
        first_date_nulls=first(const(None)) >> over(window),
    )
    result = list(query)
    expected = [
        dict(
            first_date=date(2018, 1, 1),
            last_date=date(2018, 1, 7),
            first_date_nulls=None,
        ),
        dict(
            first_date=date(2018, 1, 1),
            last_date=date(2018, 1, 7),
            first_date_nulls=None,
        ),
        dict(
            first_date=date(2018, 1, 1),
            last_date=date(2018, 1, 7),
            first_date_nulls=None,
        ),
        dict(
            first_date=date(2018, 1, 1),
            last_date=date(2018, 1, 7),
            first_date_nulls=None,
        ),
        dict(
            first_date=date(2018, 1, 2),
            last_date=date(2018, 1, 4),
            first_date_nulls=None,
        ),
        dict(
            first_date=date(2018, 1, 2),
            last_date=date(2018, 1, 4),
            first_date_nulls=None,
        ),
        dict(
            first_date=date(2018, 1, 2),
            last_date=date(2018, 1, 4),
            first_date_nulls=None,
        ),
    ]
    assert_rowset_equal(result, expected)


def test_nth(t_rows: list[dict[str, Element]]) -> None:
    query = table(t_rows) >> select(
        nth_date=nth(get("date"), const(1))
        >> over(Window.range(partition_by=[get("name")]))
    )
    result = list(query)
    expected = [
        dict(nth_date=date(2018, 1, 4)),
        dict(nth_date=date(2018, 1, 4)),
        dict(nth_date=date(2018, 1, 4)),
        dict(nth_date=date(2018, 1, 4)),
        dict(nth_date=date(2018, 1, 3)),
        dict(nth_date=date(2018, 1, 3)),
        dict(nth_date=date(2018, 1, 3)),
    ]
    assert_rowset_equal(result, expected)


def test_nth_past_frame(t_rows: list[dict[str, Element]]) -> None:
    query = table(t_rows) >> select(
        nth_date=nth(get("date"), const(4000))
        >> over(Window.range(partition_by=[get("name")]))
    )
    result = list(query)
    expected = [
        dict(nth_date=None),
        dict(nth_date=None),
        dict(nth_date=None),
        dict(nth_date=None),
        dict(nth_date=None),
        dict(nth_date=None),
        dict(nth_date=None),
    ]
    assert_rowset_equal(result, expected)


def test_nth_past_frame_preceding_following(t_rows: list[dict[str, Element]]) -> None:
    query = table(t_rows) >> select(
        nth_date=nth(get("date"), const(4000))
        >> over(
            Window.range(
                partition_by=[get("name")],
                preceding=const(200),
                following=const(1000),
            )
        )
    )
    result = list(query)
    expected = [
        dict(nth_date=None),
        dict(nth_date=None),
        dict(nth_date=None),
        dict(nth_date=None),
        dict(nth_date=None),
        dict(nth_date=None),
        dict(nth_date=None),
    ]
    assert_rowset_equal(result, expected)


def test_lead_lag(t_rows: list[dict[str, Element]]) -> None:
    window = Window.range(partition_by=[get("name")])
    query = table(t_rows) >> select(
        lead_date=lead(get("date"), const(1)) >> over(window),
        lag_date=lag(get("date"), const(1)) >> over(window),
    )
    result = list(query)
    expected: list[Mapping[str, Element]] = [
        dict(lead_date=date(2018, 1, 4), lag_date=None),
        dict(lead_date=date(2018, 1, 6), lag_date=date(2018, 1, 1)),
        dict(lead_date=date(2018, 1, 7), lag_date=date(2018, 1, 4)),
        dict(lead_date=None, lag_date=date(2018, 1, 6)),
        dict(lead_date=date(2018, 1, 3), lag_date=None),
        dict(lead_date=date(2018, 1, 4), lag_date=date(2018, 1, 2)),
        dict(lead_date=None, lag_date=date(2018, 1, 3)),
    ]
    assert_rowset_equal(result, expected)



import os
import subprocess
import sys

import pytest

from stupidb.associative.segmenttree import SegmentTree
from stupidb.functions.associative import Sum

pytestmark = pytest.mark.animate


def test_to_gif() -> None:
    from stupidb.associative.animate import SegmentTreeAnimator

    seg = SegmentTree([(i,) for i in range(1, 5)], Sum[int, int], fanout=2)
    anim = SegmentTreeAnimator(seg)
    with open(os.devnull, "wb") as io:
        anim.animate(io)


def test_main_default_args() -> None:
    try:
        subprocess.check_call(
            [f"{sys.executable}", "-m", "stupidb.associative.animate", "-o", os.devnull]
        )
    except subprocess.CalledProcessError as e:  # pragma: no cover
        print(e.stderr, file=sys.stderr)
        raise


def test_main_custom_leaves() -> None:
    try:
        subprocess.check_call(
            [
                f"{sys.executable}",
                "-m",
                "stupidb.associative.animate",
                "-o",
                os.devnull,
                "-l",
                "1",
                "-l",
                "2",
                "-f",
                "2",
            ]
        )
    except subprocess.CalledProcessError as e:  # pragma: no cover
        print(e.stderr, file=sys.stderr)
        raise



"""Navigation and simple window function interface and implementation."""

from __future__ import annotations

import operator
from typing import Callable, ClassVar, MutableMapping, Sequence

from ...typehints import Input
from .core import (
    BinaryNavigationAggregate,
    TernaryNavigationAggregate,
    UnaryNavigationAggregate,
)


class LeadLag(TernaryNavigationAggregate[Input, int, Input, Input]):
    """Base class for shifting operations.

    This class tracks the index of the current row that is being computed.

    """

    __slots__ = "index", "ninputs"
    offset_operation: ClassVar[Callable[[int, int], int]]

    @classmethod
    def offset(cls, index: int, offset: int | None) -> int:
        return -1 if offset is None else cls.offset_operation(index, offset)

    def __init__(
        self,
        inputs: Sequence[Input | None],
        offsets: Sequence[int | None],
        defaults: Sequence[Input | None],
    ) -> None:
        super().__init__(inputs, offsets, defaults)
        self.index = 0
        self.ninputs = len(inputs)

    def execute(self, begin: int, end: int) -> Input | None:
        """Compute the value of the navigation function `lead` or `lag`.

        Notes
        -----
        `begin` and `end` are ignored in lead/lag, by definition.

        """
        index = self.index
        offset = self.offset(index, self.inputs2[index])
        default = self.inputs3[index]

        # if we asked for a null offset or we're out of bounds then return a
        # null
        if offset < 0 or offset >= self.ninputs:
            result = default if default is not None else None
        else:
            result = self.inputs1[offset]

        self.index += 1
        return result


class Lead(LeadLag[Input]):
    __slots__ = ()
    offset_operation = operator.add


class Lag(LeadLag[Input]):
    __slots__ = ()
    offset_operation = operator.sub


class FirstLast(UnaryNavigationAggregate[Input, Input]):
    """Base class for first and last navigation functions.

    The difference between first and last is where the search for non NULL
    values starts.

    This aggregation keeps a cache of computed aggregations keyed by the begin
    and end of the range it's been queried over.

    """

    __slots__ = ("cache",)

    def __init__(self, inputs1: Sequence[Input | None]) -> None:
        super().__init__(inputs1)
        self.cache: MutableMapping[tuple[int, int], Input | None] = {}

    def execute(self, begin: int, end: int) -> Input | None:
        try:
            return self.cache[begin, end]
        except KeyError:
            inputs = (self.inputs1[i] for i in range(begin, end))
            value = self.cache[begin, end] = next(
                (arg for arg in inputs if arg is not None),
                None,
            )
            return value


class First(FirstLast[Input]):
    __slots__ = ()


class Last(FirstLast[Input]):
    __slots__ = ()

    def __init__(self, inputs1: Sequence[Input | None]) -> None:
        super().__init__(inputs1[::-1])


class Nth(BinaryNavigationAggregate[Input, int, Input]):
    """Compute the nth row in a window frame."""

    __slots__ = "index", "cache"

    def __init__(
        self,
        inputs1: Sequence[Input | None],
        inputs2: Sequence[int | None],
    ) -> None:
        super().__init__(inputs1, inputs2)
        self.index = 0
        self.cache: MutableMapping[tuple[int, int], Input | None] = {}

    def execute(self, begin: int, end: int) -> Input | None:
        # Assert invariants:
        # 1. The start of the range must be less than or equal to the end,
        #    which must be less than or equal to the number of input rows
        # 2. The current index must be between the begin and end of the queried
        #    range.
        assert 0 <= begin <= end <= len(self.inputs1)
        assert begin <= self.index <= end
        try:
            return self.cache[begin, end]
        except KeyError:
            # the current position in the frame
            frame_position = begin + self.index

            assert (
                frame_position <= end
            ), f"frame_position == {frame_position} :: end == {end}"

            # compute the offset relative to the current row
            offsets = self.inputs2
            target_index = offsets[frame_position]
            ninputs = end - begin

            data = self.inputs1
            if target_index is not None and -ninputs <= target_index < ninputs:
                result = data[target_index]
            else:
                # if the user asked for a row outside the frame, return
                # None
                result = None
            self.cache[begin, end] = result
        self.index += 1
        return result



import pytest

from stupidb.associative.bitset import BitSet


def test_construction() -> None:
    bs = BitSet()
    assert not bs
    assert len(bs) == 0

    bs = BitSet({1, 2})
    assert len(bs) == 2
    assert list(bs) == [1, 2]

    with pytest.raises(ValueError):
        BitSet({2, -1})


def test_repr() -> None:
    bs = BitSet()
    assert repr(bs) == "BitSet()"

    bs = BitSet([1, 2])
    assert repr(bs) == "BitSet({1, 2})"


def test_add() -> None:
    bs = BitSet()
    assert 0 not in bs

    bs.add(1)
    assert 1 in bs
    assert 2 not in bs

    bs.add(2)
    assert 1 in bs
    assert 2 in bs
    assert 0 not in bs

    with pytest.raises(ValueError):
        bs.add(-42)


def test_remove() -> None:
    bs = BitSet()
    assert 0 not in bs

    bs.add(1)
    assert 1 in bs
    assert 2 not in bs

    bs.add(2)
    assert 1 in bs
    assert 2 in bs
    assert 0 not in bs

    bs.remove(2)
    assert 1 in bs
    assert 2 not in bs

    with pytest.raises(KeyError):
        bs.remove(40)

    with pytest.raises(ValueError):
        bs.remove(-1)


def test_discard() -> None:
    bs = BitSet()
    assert 0 not in bs

    bs.add(1)
    assert 1 in bs
    assert 2 not in bs

    bs.add(2)
    assert 1 in bs
    assert 2 in bs
    assert 0 not in bs

    bs.discard(2)
    assert 1 in bs
    assert 2 not in bs

    bs.discard(40)

    with pytest.raises(ValueError):
        bs.discard(-1)


def test_iter() -> None:
    bs = BitSet()
    assert list(bs) == []

    bs.add(1)
    assert list(bs) == [1]

    bs.add(2)
    assert list(bs) == [1, 2]

    bs.add(3)
    assert list(bs) == [1, 2, 3]

    bs.remove(1)
    assert list(bs) == [2, 3]

    bs.remove(2)
    assert list(bs) == [3]


def test_len() -> None:
    bs = BitSet()
    assert len(bs) == 0

    bs.add(1)
    assert len(bs) == 1

    bs.add(2)
    assert len(bs) == 2

    bs.add(3)
    assert len(bs) == 3

    bs.remove(3)
    assert len(bs) == 2

    bs.remove(2)
    assert len(bs) == 1

    bs.remove(1)
    assert len(bs) == 0



from __future__ import annotations

from stupidb import Window, dense_rank, order_by, over, rank, row_number, select, table
from stupidb.functions.ranking import Sentinel

from .conftest import assert_rowset_equal


def test_row_number(t_rows: list[dict]) -> None:
    window = Window.range(partition_by=[lambda r: r.name])
    query = table(t_rows) >> select(row_id=row_number() >> over(window))
    result = list(query)
    expected = [
        dict(row_id=0),
        dict(row_id=1),
        dict(row_id=2),
        dict(row_id=3),
        dict(row_id=0),
        dict(row_id=1),
        dict(row_id=2),
    ]
    assert_rowset_equal(result, expected)


def test_rank() -> None:
    rows = [
        dict(name="apple"),
        dict(name="apple"),
        dict(name="grapes"),
        dict(name="grapes"),
        dict(name="orange"),
        dict(name="watermelon"),
    ]
    window = Window.rows(order_by=[lambda r: r.name])
    query = table(rows) >> select(name=lambda r: r.name, ranked=rank() >> over(window))
    result = [row.ranked for row in query]
    expected = [0, 0, 2, 2, 4, 5]
    assert result == expected


def test_rank_with_nulls() -> None:
    rows = [dict(name="a"), dict(name=None), dict(name=None), dict(name="b")]
    window = Window.rows(order_by=[lambda r: r.name])
    query = (
        table(rows)
        >> select(name=lambda r: r.name, ranked=rank() >> over(window))
        >> order_by(lambda r: r.ranked)
    )
    result = [row.ranked for row in query]
    expected = [0, 0, 2, 3]
    assert result == expected


def test_dense_rank() -> None:
    rows = [
        dict(name="apple"),
        dict(name="apple"),
        dict(name="grapes"),
        dict(name="grapes"),
        dict(name="orange"),
        dict(name="watermelon"),
    ]
    window = Window.rows(order_by=[lambda r: r.name])
    query = table(rows) >> select(
        name=lambda r: r.name, ranked=dense_rank() >> over(window)
    )
    result = [row.ranked for row in query]
    expected = [0, 0, 1, 1, 2, 3]
    assert result == expected


def test_dense_rank_with_nulls() -> None:
    rows = [dict(name="a"), dict(name=None), dict(name=None), dict(name="b")]
    window = Window.rows(order_by=[lambda r: r.name])
    query = (
        table(rows)
        >> select(name=lambda r: r.name, ranked=dense_rank() >> over(window))
        >> order_by(lambda r: r.ranked)
    )
    result = [row.ranked for row in query]
    expected = [0, 0, 1, 2]
    assert result == expected


def test_sentinel() -> None:
    sen = Sentinel()
    assert repr(sen) == "Sentinel()"
    assert sen == sen
    assert not (sen != sen)



import pytest

from stupidb.row import JoinedRow, Row


def test_row_invalid_attribute() -> None:
    row = Row({"a": 1}, _id=0)
    with pytest.raises(AttributeError):
        row.b


def test_row_equality() -> None:
    row1 = Row({"a": 1}, _id=0)
    row2 = Row({"a": 1}, _id=1)
    assert row1 == row2


def test_row_inequality() -> None:
    row1 = Row({"a": 0}, _id=0)
    row2 = Row({"a": 1}, _id=1)
    assert row1 != row2


def test_row_repr() -> None:
    row = Row({"a": 1}, _id=0)
    assert repr(row) == "Row({'a': 1})"


def test_joined_row_hash() -> None:
    row = JoinedRow({"a": 1}, {"b": 2}, _id=0)
    assert hash(row)


def test_hash_after_renew_id() -> None:
    row1 = JoinedRow({"a": 1}, {"b": 2}, _id=0)
    row2 = row1._renew_id(id=1)
    assert row1._id != row2._id
    assert hash(row1) == hash(row2)


def test_joined_row_data() -> None:
    row = JoinedRow({"a": 1}, {"b": 2}, _id=0)
    assert row.data == {"a": 1, "b": 2}


def test_joined_row_from_mapping() -> None:
    row = JoinedRow({"a": 1}, {"b": 2}, _id=0)
    with pytest.raises(TypeError):
        row.from_mapping({"c": 1})


def test_joined_row_getitem() -> None:
    row = JoinedRow({"a": 1}, {"b": 2}, _id=0)
    assert row.a == 1
    assert row.b == 2


def test_joined_row_overlapping() -> None:
    row = JoinedRow({"a": 1}, {"a": 2}, _id=0)
    with pytest.raises(ValueError):
        row.a
    assert row.left.a == 1
    assert row.right.a == 2


def test_joined_row_repr() -> None:
    row = JoinedRow({"a": 1}, {"a": 2}, _id=0)
    assert repr(row) == "JoinedRow(left={'a': 1}, right={'a': 2})"

    row = JoinedRow({"a": 1}, {"b": 2}, _id=0)
    assert repr(row) == "JoinedRow(left={'a': 1}, right={'b': 2})"



"""Algorithms for aggregation."""

from __future__ import annotations

import abc
import bisect
import enum
import functools
import typing
from typing import (
    Any,
    Callable,
    Generic,
    Iterable,
    Iterator,
    NamedTuple,
    Sequence,
    TypeVar,
)

import toolz

from .aggregator import Aggregate, Aggregator
from .functions.associative import BinaryAssociativeAggregate, UnaryAssociativeAggregate
from .functions.navigation import (
    BinaryNavigationAggregate,
    TernaryNavigationAggregate,
    UnaryNavigationAggregate,
)
from .functions.ranking import RankingAggregate
from .protocols import Comparable
from .row import AbstractRow
from .typehints import Following, OrderBy, OrderingKey, PartitionBy, Preceding, T


class StartStop(NamedTuple):
    """A class to hold start and stop values for a range of rows."""

    start: int
    stop: int


@enum.unique
class Nulls(enum.Enum):
    """An enumeration indicating how to handle null values when sorting."""

    FIRST = -1
    LAST = 1


class FrameClause(abc.ABC):
    """Class for computing frame boundaries."""

    __slots__ = "order_by", "partition_by", "preceding", "following", "nulls"

    def __init__(
        self,
        order_by: Sequence[OrderBy],
        partition_by: Sequence[PartitionBy],
        preceding: Preceding | None,
        following: Following | None,
        nulls: Nulls,
    ) -> None:
        self.order_by = order_by
        self.partition_by = partition_by
        self.preceding = preceding
        self.following = following
        self.nulls = nulls

    @abc.abstractmethod
    def find_partition_begin(
        self,
        current_row: AbstractRow,
        row_id_in_partition: int,
        current_row_order_by_value: OrderingKey | None,
        order_by_values: Sequence[OrderingKey],
    ) -> int:
        """Find the beginning of a window in a partition.

        Parameters
        ----------
        current_row
            The row relative to which we are computing the window.
        row_id_in_partition
            The zero-based index of `current_row` in possible_peers.
        current_row_order_by_value
            The value of the ORDER BY key in the current row.
        order_by_values
            The order by values for the current partition.

        Returns
        -------
        int
            The start point of the window in the current partition

        """

    @abc.abstractmethod
    def find_partition_end(
        self,
        current_row: AbstractRow,
        row_id_in_partition: int,
        current_row_order_by_value: OrderingKey | None,
        order_by_values: Sequence[OrderingKey],
    ) -> int:
        """Find the end of a window in a partition.

        Parameters
        ----------
        current_row
            The row relative to which we are computing the window.
        row_id_in_partition
            The zero-based index of `current_row` in possible_peers.
        current_row_order_by_value
            The value of the ORDER BY key in the current row.
        order_by_values
            The order by values for the current partition.

        Returns
        -------
        int
            The end point of the window in the current partition

        """

    @abc.abstractmethod
    def setup_window(
        self,
        possible_peers: Sequence[AbstractRow],
        current_row: AbstractRow,
        order_by_columns: Sequence[str],
    ) -> tuple[OrderingKey, Sequence[OrderingKey]]:
        """Compute the current row's ordering keys."""

    def compute_window_frame(
        self,
        possible_peers: Sequence[AbstractRow],
        current_row: AbstractRow,
        row_id_in_partition: int,
        order_by_columns: Sequence[str],
    ) -> StartStop:
        """Compute the bounds of the window frame.

        Parameters
        ----------
        possible_peers
            The sequence of possible rows of which the window could consist.
        current_row
            The row relative to which we are computing the window.
        row_id_in_partition
            The zero-based index of `current_row` in possible_peers.
        order_by_columns
            The columns by which we have ordered our window, if any.

        Returns
        -------
        StartStop
            The start and stop of the window frame.

        """
        current_row_order_by_value, order_by_values = self.setup_window(
            possible_peers, current_row, order_by_columns
        )

        preceding = self.preceding
        if preceding is not None:
            start = self.find_partition_begin(
                current_row,
                row_id_in_partition,
                current_row_order_by_value,
                order_by_values,
            )
        else:
            start = 0

        npeers = len(possible_peers)
        following = self.following
        if following is not None:
            stop = self.find_partition_end(
                current_row,
                row_id_in_partition,
                current_row_order_by_value,
                order_by_values,
            )
        else:
            if not all(order_by_values):
                # if we don't have an order by then all possible peers are the
                # actual peers of this row
                stop = npeers
            else:
                # default to the current row if following is not provided. This
                # is consistent with the defaults in at least PostgreSQL and
                # SQLite.
                stop = row_id_in_partition + 1

        new_start = max(start, 0)
        new_stop = min(stop, npeers)
        return StartStop(new_start, new_stop)


class RowsMode(FrameClause):
    """A frame clause implementation for window function ``ROWS`` mode.

    ``ROWS`` mode computes the window frame relative to the difference between
    the row index of the current row and what is given by ``preceding`` and
    ``following``.

    See Also
    --------
    RangeMode

    """

    __slots__ = ()

    def find_partition_begin(
        self,
        current_row: AbstractRow,
        row_id_in_partition: int,
        current_row_order_by_value: OrderingKey | None,
        order_by_values: Sequence[OrderingKey],
    ) -> int:  # noqa: D102
        preceding = self.preceding
        assert preceding is not None, "preceding is None"
        return row_id_in_partition - typing.cast(int, preceding(current_row))

    def find_partition_end(
        self,
        current_row: AbstractRow,
        row_id_in_partition: int,
        current_row_order_by_value: OrderingKey | None,
        order_by_values: Sequence[OrderingKey],
    ) -> int:  # noqa: D102
        following = self.following
        assert following is not None, "following is None"
        return row_id_in_partition + typing.cast(int, following(current_row)) + 1

    def setup_window(
        self,
        possible_peers: Sequence[AbstractRow],
        current_row: AbstractRow,
        order_by_columns: Sequence[str],
    ) -> tuple[OrderingKey, Sequence[OrderingKey]]:  # noqa: D102
        cols = [
            tuple(map(peer.__getitem__, order_by_columns)) for peer in possible_peers
        ]
        return tuple(map(current_row.__getitem__, order_by_columns)), cols


class RangeMode(FrameClause):
    """A frame clause implementation for window function ``RANGE`` mode.

    ``RANGE`` mode computes the window frame relative to the difference between
    ``preceding`` and ``following`` and the current row's ordering key.

    See Also
    --------
    RowsMode

    """

    __slots__ = ()

    def __init__(
        self,
        order_by: Sequence[OrderBy],
        partition_by: Sequence[PartitionBy],
        preceding: Preceding | None,
        following: Following | None,
        nulls: Nulls,
    ) -> None:
        n_order_by = len(order_by)
        if n_order_by > 1:
            raise ValueError(
                "Must have exactly ONE order by to use range windows. "
                f"Got {n_order_by:d} functions."
            )
        super().__init__(order_by, partition_by, preceding, following, nulls)

    def setup_window(
        self,
        possible_peers: Sequence[AbstractRow],
        current_row: AbstractRow,
        order_by_columns: Sequence[str],
    ) -> tuple[OrderingKey, Sequence[OrderingKey]]:  # noqa: D102
        # range mode allows no order by
        if not order_by_columns:
            return (), [()]

        ncolumns = len(order_by_columns)
        assert ncolumns == 1, f"ncolumns == {ncolumns:d}"
        (order_by_column,) = order_by_columns
        order_by_values = [(peer[order_by_column],) for peer in possible_peers]
        current_row_order_by_value = (current_row[order_by_column],)
        return current_row_order_by_value, order_by_values

    def find_partition_begin(
        self,
        current_row: AbstractRow,
        row_id_in_partition: int,
        current_row_order_by_values: OrderingKey | None,
        order_by_values: Sequence[OrderingKey],
    ) -> int:  # noqa: D102
        assert (
            current_row_order_by_values is not None
        ), "current_row_order_by_value is None"
        preceding = self.preceding
        assert preceding is not None, "preceding function is None"
        if not current_row_order_by_values:
            return 0
        assert len(current_row_order_by_values) == 1
        (current_row_order_by_value,) = current_row_order_by_values
        value_to_find = current_row_order_by_value - preceding(current_row)
        return bisect.bisect_left(order_by_values, (value_to_find,))

    def find_partition_end(
        self,
        current_row: AbstractRow,
        row_id_in_partition: int,
        current_row_order_by_values: OrderingKey | None,
        order_by_values: Sequence[OrderingKey],
    ) -> int:  # noqa: D102
        assert (
            current_row_order_by_values is not None
        ), "current_row_order_by_values is None"
        following = self.following
        assert following is not None, "following function is None"
        if not current_row_order_by_values:
            return len(order_by_values)
        assert len(current_row_order_by_values) == 1
        (current_row_order_by_value,) = current_row_order_by_values
        value_to_find = current_row_order_by_value + following(current_row)
        return bisect.bisect_right(order_by_values, (value_to_find,))


class Window:
    """A namespace class providing the user-facing API for windowing modes."""

    __slots__ = ()

    @staticmethod
    def rows(
        order_by: Sequence[OrderBy] = (),
        partition_by: Sequence[PartitionBy] = (),
        preceding: Preceding | None = None,
        following: Following | None = None,
        nulls: Nulls = Nulls.FIRST,
    ) -> FrameClause:
        """Construct a ``ROWS`` mode frame clause.

        ``ROWS`` windows are useful for computing over windows that can be
        determined by relative row index alone.

        See Also
        --------
        Window.range

        """
        return RowsMode(order_by, partition_by, preceding, following, nulls)

    @staticmethod
    def range(
        order_by: Sequence[OrderBy] = (),
        partition_by: Sequence[PartitionBy] = (),
        preceding: Preceding | None = None,
        following: Following | None = None,
        nulls: Nulls = Nulls.FIRST,
    ) -> FrameClause:
        """Construct a ``RANGE`` mode frame clause.

        ``RANGE`` windows can be used to compute over windows whose bounds are
        not easily determined by row number such as time based windows.

        See Also
        --------
        Window.rows

        """
        return RangeMode(order_by, partition_by, preceding, following, nulls)


Getter = Callable[[AbstractRow], Any]
ConcreteAggregate = TypeVar(
    "ConcreteAggregate",
    UnaryAssociativeAggregate,
    BinaryAssociativeAggregate,
    UnaryNavigationAggregate,
    BinaryNavigationAggregate,
    TernaryNavigationAggregate,
    RankingAggregate,
)


class AggregateSpecification(Generic[ConcreteAggregate]):
    """Specification for computing a (non-windowed) aggregation.

    Attributes
    ----------
    aggregate_type
        The aggregate class to use for aggregation.
    getters
        A tuple of callables used to produce the arguments for the aggregation.

    See Also
    --------
    WindowAggregateSpecification

    """

    __slots__ = "aggregate_type", "getters"

    def __init__(
        self,
        aggregate_type: type[ConcreteAggregate],
        *getters: Getter,
    ) -> None:
        self.aggregate_type: type[ConcreteAggregate] = aggregate_type
        self.getters = getters


def row_key_compare(
    order_func: Callable[[AbstractRow], tuple[Comparable[T], ...]],
    null_ordering: Nulls,
    left_row: AbstractRow,
    right_row: AbstractRow,
) -> int:
    """Compare `left_row` and `right_row` using `order_by`.

    Notes
    -----
    ``NULL`` ordering is handled using `null_ordering`.

    """
    for left_key, right_key in zip(order_func(left_row), order_func(right_row)):
        if left_key is None and right_key is not None:
            return null_ordering.value
        if left_key is not None and right_key is None:
            return -null_ordering.value
        if left_key is None and right_key is None:
            return 0

        assert left_key is not None, "left_key is None"
        assert right_key is not None, "right_key is None"
        if left_key < right_key:
            return -1
        if left_key > right_key:
            return 1
    return 0


def make_key_func(
    order_func: Callable[[AbstractRow], tuple[Comparable[T], ...]],
    nulls: Nulls,
) -> Callable[[AbstractRow], OrderingKey[T]]:
    """Make a function usable with the key argument to sorting functions.

    This return value of this function can be passed to
    :func:`sorted`/:meth:`list.sort`.

    Parameters
    ----------
    order_by_columns
        A sequence of :class:`str` instances referring to the keys of an
        :class:`~stupidb.row.AbstractRow`.

    """
    return functools.cmp_to_key(functools.partial(row_key_compare, order_func, nulls))


class WindowAggregateSpecification(Generic[ConcreteAggregate]):
    """A specification for a window aggregate.

    Attributes
    ----------
    aggregate_type
        The class of :data:`~stupidb.aggregation.ConcreteAggregate` to use for
        aggregation.
    getters
        A tuple of functions that produce single column values given an
        instance of :class:`~stupidb.row.AbstractRow`.
    frame_clause
        A thin struct encapsulating the details of the window such as ``ORDER
        BY`` (:attr:`stupidb.aggregation.FrameClause.order_by`), ``PARTITION
        BY`` (:attr:`stupidb.aggregation.FrameClause.partition_by`) and
        preceding and following.

    See Also
    --------
    stupidb.aggregation.FrameClause

    """

    __slots__ = "aggregate_type", "getters", "frame_clause"

    def __init__(
        self,
        aggregate_type: type[ConcreteAggregate],
        getters: tuple[Getter, ...],
        frame_clause: FrameClause,
    ) -> None:
        self.aggregate_type: type[ConcreteAggregate] = aggregate_type
        self.getters = getters
        self.frame_clause = frame_clause

    def compute(self, rows: Iterable[AbstractRow]) -> Iterator[T | None]:
        """Aggregate `rows` over a window, producing an iterator of results.

        Parameters
        ----------
        rows
            An :class:`~typing.Iterable` of rows.

        """
        frame_clause = self.frame_clause
        order_by = frame_clause.order_by

        # Generate names for temporary order by columns, users never see these.
        #
        # TODO: If we had static schema information these wouldn't be necessary
        # in cases where the ordering keys are named columns (either physical
        # or computed)
        order_by_columns = [f"_order_by_{i:d}" for i in range(len(order_by))]

        # Add computed order by columns that are used when evaluating window
        # functions in range mode
        # TODO: check that if in range mode we only have single order by
        order_func = toolz.juxt(*order_by)
        rows_for_partition = (
            row.merge(dict(zip(order_by_columns, order_func(row)))) for row in rows
        )

        # divide the input rows into partitions
        #
        # we only need the values once we've grouped, so there's no need to
        # store a possibly expensive key like a tuple, so we store just the
        # integer result of hash in the dict
        #
        # we also only need the partition values once the rows have
        # been partitioned
        partitions = toolz.groupby(
            toolz.compose(hash, toolz.juxt(*frame_clause.partition_by)),
            rows_for_partition,
        ).values()

        # aggregation results, preallocated to avoid the need to sort
        # before returning: we later assign elements to this list using
        # the original row id
        results: list[T | None] = [None] * sum(map(len, partitions))

        # Aggregate over each partition
        aggregate_type = self.aggregate_type
        getters = self.getters
        key_func = make_key_func(order_func, frame_clause.nulls)
        for possible_peers in partitions:
            # sort the partition according to the ordering key
            possible_peers.sort(key=key_func)

            # Construct an aggregator for the function being computed
            #
            # For navigation functions like lead, lag, first, last and nth, we
            # construct a simple structure that computes the current value
            # of the navigation function given the inputs. We use the same
            # approach for ranking functions such as row_number, rank, and
            # dense_rank.
            #
            # For associative aggregations we construct a segment tree
            # using `arguments` as the leaves, with `aggregate` instances as
            # the interior nodes. Each node (both leaves and non-leaves) is a
            # state of the aggregation. The leaves are the initial states, the
            # root is the final state.
            aggregator: Aggregator[Aggregate, T] = aggregate_type.prepare(
                possible_peers, getters, order_by_columns
            )

            # For every row in the set of possible peers of the current row
            # compute the window frame, and query the aggregator for the value
            # of the aggregation within that frame.
            for row_id_in_partition, row in enumerate(possible_peers):
                start, stop = frame_clause.compute_window_frame(
                    possible_peers, row, row_id_in_partition, order_by_columns
                )
                # Assign the result to the position of the original row id
                # because we processed them in partition order, which might not
                # be the same as the input order.
                results[row._id] = aggregator.query(start, stop)

        return iter(results)



from __future__ import annotations

from stupidb.protocols import Comparable


class CustomComparable(Comparable):
    def __init__(self, value: float) -> None:
        self.value = value

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CustomComparable):
            return NotImplemented
        return self.value == other.value

    def __lt__(self, other: CustomComparable) -> bool:
        return self.value < other.value


def test_comparable_implements_methods_eq() -> None:
    comp = CustomComparable(1.0)
    comp2 = CustomComparable(1.0)
    assert comp == comp2
    assert not (comp != comp2)
    assert not (comp < comp2)
    assert not (comp > comp2)
    assert comp <= comp2
    assert comp >= comp2


def test_comparable_implements_methods_ne() -> None:
    comp = CustomComparable(1.0)
    comp2 = CustomComparable(2.0)
    assert comp != comp2
    assert comp < comp2
    assert comp <= comp2
    assert not (comp == comp2)
    assert not (comp > comp2)
    assert not (comp >= comp2)



from stupidb.api import (
    difference,
    difference_all,
    intersect,
    intersect_all,
    order_by,
    union,
    union_all,
)

from .conftest import assert_rowset_equal


def test_union_distinct() -> None:
    rows = [dict(name="a"), dict(name="b")]
    other_rows = [dict(name="c")]
    query = rows >> union(other_rows) >> order_by(lambda r: r.name)
    result = list(query)
    expected = [dict(name="a"), dict(name="b"), dict(name="c")]
    assert result == expected


def test_union_duplicates() -> None:
    rows = [dict(name="a"), dict(name="b")]
    other_rows = [dict(name="a")]
    query = rows >> union(other_rows) >> order_by(lambda r: r.name)
    result = list(query)
    expected = [dict(name="a"), dict(name="b")]
    assert result == expected


def test_union_all_distinct() -> None:
    rows = [dict(name="a"), dict(name="b")]
    other_rows = [dict(name="c")]
    query = rows >> union_all(other_rows) >> order_by(lambda r: r.name)
    result = list(query)
    expected = [dict(name="a"), dict(name="b"), dict(name="c")]
    assert result == expected


def test_union_all_duplicates() -> None:
    rows = [dict(name="a"), dict(name="b")]
    other_rows = [dict(name="a")]
    query = rows >> union_all(other_rows) >> order_by(lambda r: r.name)
    result = list(query)
    expected = [dict(name="a"), dict(name="a"), dict(name="b")]
    assert result == expected


def test_intersect_distinct() -> None:
    rows = [dict(name="a"), dict(name="b")]
    other_rows = [dict(name="c")]
    query = rows >> intersect(other_rows) >> order_by(lambda r: r.name)
    result = list(query)
    expected: list[dict[str, str]] = []
    assert result == expected


def test_intersect_duplicates() -> None:
    rows = [dict(name="a"), dict(name="b")]
    other_rows = [dict(name="a")]
    query = rows >> intersect(other_rows) >> order_by(lambda r: r.name)
    result = list(query)
    expected = [dict(name="a")]
    assert result == expected


def test_intersect_all_distinct() -> None:
    rows = [dict(name="a"), dict(name="b")]
    other_rows = [dict(name="c")]
    query = rows >> intersect_all(other_rows) >> order_by(lambda r: r.name)
    result = list(query)
    expected: list[dict[str, str]] = []
    assert result == expected


def test_intersect_all_duplicates() -> None:
    rows = [dict(name="a"), dict(name="b")]
    other_rows = [dict(name="a")]
    query = rows >> intersect_all(other_rows) >> order_by(lambda r: r.name)
    result = list(query)
    expected = [dict(name="a"), dict(name="a")]
    assert result == expected


def test_difference_distinct() -> None:
    rows = [dict(name="a"), dict(name="b")]
    other_rows = [dict(name="c")]
    query = rows >> difference(other_rows) >> order_by(lambda r: r.name)
    result = list(query)
    expected = rows[:]
    assert result == expected


def test_difference_duplicates() -> None:
    rows = [dict(name="a"), dict(name="b")]
    other_rows = [dict(name="a")]
    query = rows >> difference(other_rows) >> order_by(lambda r: r.name)
    result = list(query)
    expected = [dict(name="b")]
    assert result == expected
    assert_rowset_equal(result, expected)


def test_difference_all_distinct() -> None:
    rows = [dict(name="a"), dict(name="b"), dict(name="a")]
    other_rows = [dict(name="c"), dict(name="b")]
    query = rows >> difference_all(other_rows) >> order_by(lambda r: r.name)
    result = list(query)
    expected = [dict(name="a"), dict(name="a")]
    assert result == expected


def test_difference_all_duplicates() -> None:
    rows = [dict(name="a"), dict(name="b"), dict(name="b")]
    other_rows = [dict(name="a")]
    query = rows >> difference_all(other_rows) >> order_by(lambda r: r.name)
    result = list(query)
    expected = [dict(name="b"), dict(name="b")]
    assert result == expected



"""Module containing classes for representing rows."""

from __future__ import annotations

import abc
from typing import Any, Hashable, Iterator, Mapping

import toolz


class AbstractRow(Mapping[str, Any], Hashable, abc.ABC):
    """The base immutable row type of StupidDB.

    This is the primitive immutable tuple type of the objects that are in
    user facing APIs. They behave nearly identically to a standard
    :class:`typing.Mapping`, with the exception that they are hashable and
    values can be accessed with square-bracket syntax as well as dot notation.

    Attributes
    ----------
    pieces
        One or more mappings that make up row. One for most relations, and two
        for joins.
    _id
        The index of this row in a table. This a private field whose details
        are subject to change without notice.
    _hash
        The hash of the row's data. This stored on the instance to avoid
        recomputation in :class:`~stupidb.stupidb.SetOperation` instances, for
        example.

    """

    __slots__ = "pieces", "_id", "_hash"

    def __init__(
        self,
        piece: Mapping[str, Any],
        *pieces: Mapping[str, Any],
        _id: int = -1,
        _hash: int | None = None,
    ) -> None:
        """Construct an :class:`AbstractRow`.

        Parameters
        ----------
        piece
            A mapping from :class:`str` to :class:`typing.Any`.
        pieces
            A tuple of mappings from :class:`str` to :class:`typing.Any`.

        """
        self.pieces = piece, *pieces
        self._id = _id
        self._hash = (
            _hash
            if _hash is not None
            else hash(
                tuple(tuple(item) for piece in self.pieces for item in piece.items())
            )
        )

    def __hash__(self) -> int:
        return self._hash

    def __eq__(self, other: Any) -> bool:
        if isinstance(self, AbstractRow) and isinstance(other, AbstractRow):
            return hash(self) == hash(other)
        return getattr(self, "data", self) == other

    def __ne__(self, other: Any) -> bool:
        return not (self == other)

    @property
    @abc.abstractmethod
    def data(self) -> Mapping[str, Any]:
        """Return the underlying data for this row."""

    def __iter__(self) -> Iterator[str]:
        return iter(self.data)

    def __len__(self) -> int:
        """Return the number of columns in this row."""
        return len(self.data)

    def __getattr__(self, attr: str) -> Any:
        try:
            return self[attr]
        except KeyError as e:
            raise AttributeError(attr) from e

    def __getitem__(self, column: str) -> Any:
        return self.data[column]

    def _renew_id(self, id: int) -> AbstractRow:
        """Reify this row with a new id `_id`.

        Parameters
        ----------
        id
            The return value's new `_id`.

        """
        if self._id == id:
            # if the id is already the same as the requested id,
            # don't make a new row
            return self
        return type(self)(*self.pieces, _id=id, _hash=self._hash)


class Row(AbstractRow):
    """A concrete :class:`AbstractRow` subclass for single child relations."""

    def merge(self, other: Mapping[str, Any]) -> Row:
        """Combine the :class:`typing.Mapping` `other` with this one.

        Parameters
        ----------
        other
            Any Mapping whose keys are instances of :class:`str`.

        """
        return type(self)(
            toolz.merge(self.data, getattr(other, "data", other)), _id=self._id
        )

    @property
    def data(self) -> Mapping[str, Any]:
        """Return the underlying mapping of this :class:`Row`."""
        return self.pieces[0]

    @classmethod
    def from_mapping(cls, mapping: Mapping[str, Any], *, _id: int = -1) -> AbstractRow:
        """Construct a Row instance from any mapping with string keys.

        Parameters
        ----------
        mapping
            Any mapping with :class:`str` keys.
        _id
            A new row id for the returned :class:`Row` instance.

        """
        return (
            mapping._renew_id(_id)
            if isinstance(mapping, AbstractRow)
            else cls(getattr(mapping, "data", mapping), _id=_id)
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.data})"


class JoinedRow(AbstractRow):
    """A concrete :class:`AbstractRow` subclass for two-child relations.

    This row type is used to represent rows of a relation with two children.
    Currently this is only used for :class:`~stupidb.stupidb.Join` relations.

    .. note::

       :class:`JoinedRow` is the row type yielded when iterating over an
       instance of :class:`~stupidb.stupidb.Join`.  If you want to consume the
       rows of a join and there are overlapping column names in the left and
       right relations you must select from the :attr:`left` and :attr:`right`
       attributes of instances of this class to disambiguate.

    Attributes
    ----------
    left
        A row from the left relation
    right
        A row from the right relation

    """

    def __init__(
        self,
        left: Mapping[str, Any],
        right: Mapping[str, Any],
        *,
        _id: int = -1,
        _hash: None | None = None,
    ) -> None:
        """Construct a :class:`JoinedRow` instance.

        Parameters
        ----------
        left
            A mapping of :class:`str` to :class:`typing.Any`.
        right
            A mapping of :class:`str` to :class:`typing.Any`.
        _id
            The row id of this row.

        """
        self.left = Row.from_mapping(left, _id=_id)
        self.right = Row.from_mapping(right, _id=_id)
        self._overlapping_keys = left.keys() & right.keys()
        self._data = toolz.merge(left, right)
        super().__init__(left, right, _id=_id, _hash=_hash)

    @property
    def data(self) -> Mapping[str, Any]:
        """Return the underlying data of the row."""
        return self._data

    @classmethod
    def from_mapping(cls, *args: Any, **kwargs: Any) -> JoinedRow:
        """Raise an error.

        A :class:`JoinedRow` cannot be constructed from an arbitrary
        :class:`typing.Mapping`.

        """
        raise TypeError(f"`from_mapping` not supported for {cls.__name__!r}")

    def __getitem__(self, key: str) -> Any:
        overlap = self._overlapping_keys
        if overlap:
            raise ValueError(
                f"Joined rows have overlapping columns: {overlap!r}. "
                "Use `row.left` or `row.right` to choose the appropriate key."
            )
        return super().__getitem__(key)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(left={self.left.data}, "
            f"right={self.right.data})"
        )



from __future__ import annotations

import abc
from typing import Generic, Sequence

from ...aggregator import Aggregate, Aggregator
from ...typehints import Input1, Input2, Input3, Output, Result, T


class NavigationAggregator(Aggregator["NavigationAggregate", Result]):
    """Custom aggregator for navigation functions.

    This aggregator is useful for a subset of window functions whose underlying
    binary combine operator (if it even exists) is not associative or easy to
    express without special knowledge of the underlying aggregator
    representation.

    See Also
    --------
    stupidb.ranking.RankingAggregator

    """

    __slots__ = ("aggregate",)

    def __init__(
        self,
        inputs: Sequence[tuple[T, ...]],
        aggregate_type: type[NavigationAggregate],
    ) -> None:
        self.aggregate = aggregate_type(*zip(*inputs))

    def query(self, begin: int, end: int) -> Result | None:
        return self.aggregate.execute(begin, end)


class NavigationAggregate(Aggregate[Output]):
    """Base class for navigation aggregate functions."""

    __slots__ = ()

    @abc.abstractmethod
    def execute(self, begin: int, end: int) -> Output | None:
        """Execute the aggregation over the range from `begin` to `end`."""

    @classmethod
    def aggregator_class(cls, inputs: Sequence[T]) -> NavigationAggregator:
        return NavigationAggregator(inputs, cls)


class UnaryNavigationAggregate(Generic[Input1, Output], NavigationAggregate[Output]):
    """Navigation function taking one argument."""

    __slots__ = ("inputs1",)

    def __init__(self, inputs1: Sequence[Input1 | None]) -> None:
        self.inputs1 = inputs1


class BinaryNavigationAggregate(
    Generic[Input1, Input2, Output], NavigationAggregate[Output]
):
    """Navigation function taking two arguments."""

    __slots__ = "inputs1", "inputs2"

    def __init__(
        self,
        inputs1: Sequence[Input1 | None],
        inputs2: Sequence[Input2 | None],
    ) -> None:
        self.inputs1 = inputs1
        self.inputs2 = inputs2


class TernaryNavigationAggregate(
    Generic[Input1, Input2, Input3, Output], NavigationAggregate[Output]
):
    """Navigation function taking three arguments."""

    __slots__ = "inputs1", "inputs2", "inputs3"

    def __init__(
        self,
        inputs1: Sequence[Input1 | None],
        inputs2: Sequence[Input2 | None],
        inputs3: Sequence[Input3 | None],
    ) -> None:
        self.inputs1 = inputs1
        self.inputs2 = inputs2
        self.inputs3 = inputs3



from __future__ import annotations

import abc
from typing import Sequence

from ...aggregator import Aggregate, Aggregator
from ...protocols import Comparable
from ...row import AbstractRow
from ...typehints import Getter, Output, Result, T


class RankingAggregator(Aggregator["RankingAggregate", Result]):
    """Custom aggregator for ranking window functions.

    This aggregator is required for ranking functions because ranking functions
    take no arguments, but need access to the rows produced by the ordering
    key.

    See Also
    --------
    stupidb.navigation.NavigationAggregator
    stupidb.associative.SegmentTree

    """

    __slots__ = ("aggregate",)

    def __init__(
        self,
        order_by_values: Sequence[tuple[Comparable | None, ...]],
        aggregate_type: type[RankingAggregate],
    ) -> None:
        self.aggregate = aggregate_type(order_by_values)

    def query(self, begin: int, end: int) -> Result | None:
        """Compute the aggregation over the range of rows between `begin` and `end`."""
        return self.aggregate.execute(begin, end)


class RankingAggregate(Aggregate[Output]):
    """Base ranking aggregation class."""

    __slots__ = ("order_by_values",)

    def __init__(
        self, order_by_values: Sequence[tuple[Comparable | None, ...]]
    ) -> None:
        super().__init__()
        self.order_by_values = order_by_values

    @abc.abstractmethod
    def execute(self, begin: int, end: int) -> Output | None:
        """Compute an abstract row rank value for rows between `begin` and `end`."""

    @classmethod
    def prepare(
        cls,
        possible_peers: Sequence[AbstractRow],
        getters: tuple[Getter, ...],
        order_by_columns: Sequence[str],
    ) -> RankingAggregator[Output]:
        """Construct the aggregator for ranking."""
        order_by_values = [
            tuple(peer[column] for column in order_by_columns)
            for peer in possible_peers
        ]
        return cls.aggregator_class(order_by_values)

    @classmethod
    def aggregator_class(
        cls, values: Sequence[tuple[T | None, ...]]
    ) -> RankingAggregator[Output]:
        return RankingAggregator(values, cls)
